#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains form widgets for library
"""

from __future__ import print_function, division, absolute_import

import re
import sys
import traceback
from functools import partial

from Qt.QtCore import *
from Qt.QtWidgets import *
from Qt.QtGui import *

from tpDcc.libs import qt
from tpDcc.libs.python import decorators
from tpDcc.libs.qt.widgets import label, color


class FormDialog(QFrame, object):

    accepted = Signal(object)
    rejected = Signal(object)

    def __init__(self, parent=None, form=None):
        super(FormDialog, self).__init__(parent)

        self._settings = None

        self.main_layout = QVBoxLayout(self)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.setSpacing(0)
        self.setLayout(self.main_layout)

        self._widgets = list()
        self._validator = None

        self._title = QLabel(self)
        self._title.setObjectName('title')
        self._title.setText('FORM')
        self._description = QLabel(self)
        self._description.setObjectName('description')
        self._form_widget = FormWidget(self)
        self._form_widget.setObjectName('formWidget')
        self._form_widget.validated.connect(self._on_validated)
        btn_layout = QHBoxLayout(self)
        btn_layout.setContentsMargins(0, 0, 0, 0)
        btn_layout.setSpacing(0)
        self._accept_btn = QPushButton(self)
        self._accept_btn.setObjectName('acceptButton')
        self._accept_btn.setText('Accept')
        self._accept_btn.clicked.connect(self.accept)
        self._reject_btn = QPushButton(self)
        self._reject_btn.setObjectName('rejectButton')
        self._reject_btn.setText('Cancel')
        self._reject_btn.clicked.connect(self.reject)
        btn_layout.addSpacerItem(QSpacerItem(10, 0, QSizePolicy.Expanding, QSizePolicy.Preferred))
        btn_layout.addWidget(self._accept_btn)
        btn_layout.addWidget(self._reject_btn)

        self.main_layout.addWidget(self._title)
        self.main_layout.addWidget(self._description)
        self.main_layout.addWidget(self._form_widget)
        self.main_layout.addStretch(1)
        self.main_layout.addLayout(btn_layout)

        if form:
            self.set_settings(form)

    def accept_button(self):
        """
        Returns the accept button
        :return: QPushButton
        """

        return self._accept_btn

    def reject_button(self):
        """
        Returns the reject button
        :return: QPushButton
        """

        return self._reject_btn

    def set_settings(self, settings):
        """
        Sets dialog form settings
        :param settings: dict
        """

        self._settings = settings
        title = settings.get("title")
        if title is not None:
            self._title.setText(title)
        callback = settings.get("accepted")
        if not callback:
            self._settings["accepted"] = self._validate_accepted
        callback = settings.get("rejected")
        if not callback:
            self._settings["rejected"] = self._validate_rejected
        description = settings.get("description")
        if description is not None:
            self._description.setText(description)
        validator = settings.get("validator")
        if validator is not None:
            self._form_widget.set_validator(validator)
        layout = settings.get("layout")
        schema = settings.get("schema")
        if schema is not None:
            self._form_widget.set_schema(schema, layout=layout)

    def accept(self):
        """
        Function called when the dialog is accepted
        """

        callback = self._settings.get('accepted')
        if callback:
            callback(**self._form_widget.values())
        self.close()

    def reject(self):
        """
        Function called when the dialog is rejected
        """

        callback = self._settings.get('rejected')
        if callback:
            callback(**self._form_widget.default_values())
        self.close()

    def _validate_accepted(self, **kwargs):
        """
        Internal function called when the accept button has been clicked
        :param kwargs: dict, default values of the form fields
        """

        self._form_widget.validator()(**kwargs)

    def _validate_rejected(self, **kwargs):
        """
        Internal function called when reject button has been clicked
        :param kwargs: dict, default values of the form fields
        """

        self._form_widget.validator()(**kwargs)

    def _on_validated(self):
        """
        Internal callback function that is triggered when the has been validated
        """

        self._accept_btn.setEnabled(not self._form_widget.has_errors())


class FormWidget(QFrame, object):

    accepted = Signal(object)
    stateChanged = Signal()
    validated = Signal()

    def __init__(self, *args, **kwargs):
        super(FormWidget, self).__init__(*args, **kwargs)

        self._schema = list()
        self._widgets = list()
        self._validator = None

        main_layout = QVBoxLayout(self)
        main_layout.setContentsMargins(0, 0, 0, 0)
        main_layout.setSpacing(0)
        self.setLayout(main_layout)

        self._options_frame = QFrame(self)
        self._options_frame.setObjectName('optionsFrame')
        options_layout = QVBoxLayout(self._options_frame)
        options_layout.setContentsMargins(0, 0, 0, 0)
        options_layout.setSpacing(0)
        self._options_frame.setLayout(options_layout)

        self._title_widget = QPushButton(self)
        self._title_widget.setCheckable(True)
        self._title_widget.setObjectName('titleWidget')
        self._title_widget.toggled.connect(self._on_title_clicked)
        self._title_widget.hide()

        main_layout.addWidget(self._title_widget)
        main_layout.addWidget(self._options_frame)

    def title_widget(self):
        """
        Returns the title widget
        :return: QWidget
        """

        return self._title_widget

    def set_title(self, title):
        """
        Sets the title text
        :param title: str
        """

        self.title_widget().setText(title)

    def is_expanded(self):
        """
        Returns whether the item is expanded or not
        :return: bool
        """

        return self._title_widget.isChecked()

    def set_expanded(self, flag):
        """
        Expands the options if True, otherwise collapses the options
        :param flag: bool
        """

        self._title_widget.blockSignals(True)
        try:
            self._title_widget.setChecked(flag)
            self._options_frame.setVisible(flag)
        finally:
            self._title_widget.blockSignals(False)

    def set_title_visible(self, flag):
        """
        Sets whether the title widget is visible or not
        :param flag: bool
        """

        self.title_widget().setVisible(flag)

    def widget(self, name):
        """
        Returns the widget for the given widget name
        :param name: str
        :return: FieldWidget
        """

        for widget in self._widgets:
            if widget.data().get('name') == name:
                return widget

    def value(self, name):
        """
        Returns the value for the given widget name
        :param name: str
        :return: object
        """

        widget = self.widget(name)
        return widget.value()

    def set_value(self, name, value):
        """
        Sets the value for the given field name
        :param name: str
        :param value: variant
        """

        widget = self.widget(name)
        widget.set_value(value)

    def values(self):
        """
        Returns all the field values indexed by the field name
        :return: dict
        """

        values = dict()
        for widget in self._widgets:
            values[widget.data().get('name')] = widget.value()

        return values

    def default_values(self):
        """
        Returns all teh default field values indexed by the field name
        :return: dict
        """

        values = dict()
        for widget in self._widgets:
            values[widget.data().get('name')] = widget.default()

        return values

    def options(self):
        """
        Returns fields options
        :return: list(dict)
        """

        options = list()
        for widget in self._widgets:
            options.append(widget.data())

        return options

    def state(self):
        """
        Returns the current state
        :return: dict
        """

        options = list()
        for widget in self._widgets:
            options.append(widget.state())

        state = {
            'options': options,
            'expanded': self.is_expanded()
        }

        return state

    def set_state(self, state):
        """
        Sets the current state
        :param state: dict
        """

        expanded = state.get('expanded')
        if expanded is not None:
            self.set_expanded(expanded)

        options = state.get('options')
        if options is not None:
            self._set_state(options)

        self.validate()

    def options_state(self):
        """
        Returns options state
        :return: dict
        """

        state = dict()
        values = self.values()
        options = self.options()
        for option in options:
            name = option.get('name')
            persistent = option.get('persistent')
            if name and persistent:
                state[name] = values[name]

        return state

    def set_state_from_options(self, options):
        """
        Sets state from given options
        :param options: list(dict)
        """

        state = list()
        for option in options:
            state.append({'name': option, 'value': options[option]})

        self._set_state(state)

    def set_schema(self, schema, layout=None):
        """
        Sets the schema for the widget
        :param schema: ilst(dict)
        :param layout: str
        """

        self._schema = schema
        for data in schema:
            cls = FIELD_WIDGET_REGISTRY.get(data.get('type', 'label'))
            if not cls:
                qt.logger.warning('Cannot find widget for {}'.format(data))
                continue
            if layout and not data.get('layout'):
                data['layout'] = layout

            widget = cls(data=data)
            widget.set_data(data)

            value = data.get('value')
            default = data.get('default')
            if value is None and default is not None:
                widget.set_value(default)

            self._widgets.append(widget)

            callback = partial(self._on_option_changed, widget)
            widget.valueChanged.connect(callback)

            self._options_frame.layout().addWidget(widget)

    def validator(self):
        """
        Returns the validator for the form
        :return: fn
        """

        return self._validator

    def set_validator(self, validator):
        """
        Sets the validator for the options
        :param validator: fn
        """

        self._validator = validator

    def reset(self):
        """
        Reset all option widget back to the ir default values
        """

        for widget in self._widgets:
            widget.reset()
        self.validate()

    def validate(self):
        """
        Validates the current options using the validator
        """

        if self._validator:
            fields = self._validator(**self.values())
            if fields:
                self._set_state(fields)
            self.validated.emit()
        else:
            qt.logger.debug('No validator set')

    def has_errors(self):
        """
        Returns whether the form contains any error
        :return: bool
        """

        for widget in self._widgets:
            if widget.data().get('error'):
                return True

        return False

    def _set_state(self, fields):
        """
        Internal function that sets fields state
        :param fields: list(dict)
        """

        for widget in self._widgets:
            widget.blockSignals(True)

        try:
            for widget in self._widgets:
                widget.set_error('')
                for field in fields:
                    if field.get('name') == widget.data().get('name'):
                        widget.set_data(field)
        finally:
            for widget in self._widgets:
                widget.blockSignals(False)

        self.stateChanged.emit()

    def _on_title_clicked(self, toggle):
        """
        Internal callback function that is triggered when the user clicks in the title widget
        """

        self.set_expanded(toggle)
        self.stateChanged.emit()

    def _on_option_changed(self, widget):
        """
        Internal callback function triggered when the given option widget changes its value
        :param widget: FieldWidget
        """

        self.validate()


class FieldWidget(QFrame, object):

    valueChanged = Signal()

    DefaultLayout = 'horizontal'

    def __init__(self, parent=None, data=None):
        super(FieldWidget, self).__init__(parent)

        self._data = data or dict()
        self._error = False
        self._widget = None
        self._default = None
        self._required = None
        self._error_label = None
        self._menu_button = None
        self._action_result = None

        self.setObjectName('fieldWidget')

        direction = self._data.get('layout', self.DefaultLayout)
        if direction == 'vertical':
            main_layout = QVBoxLayout(self)
        else:
            main_layout = QHBoxLayout(self)
        main_layout.setContentsMargins(0, 0, 0, 0)
        main_layout.setSpacing(0)
        self.setLayout(main_layout)

        self._label = QLabel(self)
        self._label.setObjectName('label')
        self._label.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        main_layout.addWidget(self._label)

        self._layout2 = QHBoxLayout(self)
        main_layout.addLayout(self._layout2)
        if direction == 'vertical':
            self._label.setAlignment(Qt.AlignLeft | Qt .AlignVCenter)
        else:
            self._label.setAlignment(Qt.AlignRight | Qt.AlignVCenter)

    @staticmethod
    def to_title(name):
        """
        Converts camel case strings to title strings
        :param name: str
        :return: str
        """

        s1 = re.sub('(.)([A-Z][a-z]+)', r'\1 \2', name)
        return re.sub('([a-z0-9])([A-Z])', r'\1 \2', s1).title()

    @decorators.abstractmethod
    def value(self):
        """
        Returns the current value fo the field widget
        :return: variant
        """

        raise NotImplementedError('value function of FieldWidget is not implemented!')

    @decorators.abstractmethod
    def set_items(self, items):
        """
        Sets the items for the field widget
        :param items: list(str)
        """

        raise NotImplementedError('set_items function of FieldWidget is not implemented!')

    def label(self):
        """
        Returns label widget
        :return: QLabel
        """

        return self._label

    def title(self):
        """
        Returns the title to be displayed for the field
        :return: str
        """

        data = self.data()
        title = data.get('title', '') or data.get('name', '')
        if title:
            title = self.to_title(title)

        if self.is_required():
            title += '*'

        return title

    def is_default(self):
        """
        Returns whether the current value is the same as the default value or not
        :return: bool
        """

        return self.value() == self.default()

    def default(self):
        """
        Returns the default value for the field widget
        :return: variant
        """

        return self._default

    def set_default(self, default):
        """
        Sets teh default value for the field widget
        :param default: variant
        """

        self._default = default

    def set_text(self, text):
        """
        Sets the label text for the field widget
        :param text: str
        """

        self._label.setText(text)

    def set_value(self, value):
        """
        Sets the value of the field widget
        :param value: variant
        """

        self._on_emit_value_changed()

    def state(self):
        """
        Returns the current state of the data
        :return: dict
        """

        return {
            'name': self._data['name'],
            'value': self.value()
        }

    def data(self):
        """
        Returns the data for the widget
        :return: dict
        """

        return self._data

    def set_data(self, data):
        """
        Sets the current state of the field widget using a dictionary
        :param data: dict
        """

        state = data
        self.blockSignals(True)

        try:
            items = state.get('items', None)
            if items is not None:
                self.set_items(items)
            value = state.get('value', None)
            default = state.get('default', None)
            if default is not None:
                self.set_default(default)
            elif value is not None:
                self.set_default(value)
            if value is not None and value != self.value():
                try:
                    self.set_value(value)
                except TypeError as e:
                    qt.logger.exception('{} | {}'.format(e, traceback.format_exc()))
            enabled = state.get('enabled', None)
            if enabled is not None:
                self.setEnabled(enabled)
                self._label.setEnabled(enabled)
            hidden = state.get('hidden', None)
            if hidden is not None:
                self.setHidden(hidden)
            required = state.get('required', None)
            if required is not None:
                self.set_required(required)
            error = state.get('error', None)
            if error is not None:
                self.set_error(error)
            tooltip = state.get('toolTip', None)
            if tooltip is not None:
                self.setToolTip(tooltip)
                self.setStatusTip(tooltip)
            style = state.get('style', None)
            if style is not None:
                self.setStyleSheet(style)
            title = self.title() or ''
            self.set_text(title)
            lbl = state.get('label')
            if lbl is not None:
                text = lbl.get('name', None)
                if text is not None:
                    self.set_text(text)
                visible = lbl.get('visible', None)
                if visible is not None:
                    self.label().setVisible(visible)
            actions = state.get('actions', None)
            if actions is not None:
                self._menu_button.setVisible(True)
            menu = state.get('menu', None)
            if menu is not None:
                text = menu.get('name')
                if text is not None:
                    self._menu_button.setText(text)
                visible = menu.get('visible', True)
                self._menu_button.setVisible(visible)
            self._data.update(data)
            self.refresh()
        finally:
            self.blockSignals(False)

    def set_error(self, message):
        """
        Sets the error message to be displayed for the field widget
        :param message: str
        """

        self._error = True if message else False
        self._data['error'] = message
        if self._error:
            self._error_label.setText(message)
            self._error_label.setHidden(False)
            self.setToolTip(message)
        else:
            self._error_label.setText('')
            self._error_label.setHidden(True)
            self.setToolTip(self.data().get('toolTip'))

        self.refresh()

    def widget(self):
        """
        Returns the widget used to set and get the field value
        :return: QWidget
        """

        return self._widget

    def set_widget(self, widget):
        """
        Sets the widget used to set and get the field value
        :param widget: QWidget
        """

        widget_layout = QHBoxLayout()
        widget_layout.setContentsMargins(0, 0, 0, 0)
        widget_layout.setSpacing(0)

        self._widget = widget
        self._widget.setParent(self)
        self._widget.setObjectName('widget')
        self._widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)

        self._menu_button = QPushButton('...')
        self._menu_button.setHidden(True)
        self._menu_button.setObjectName('menuButton')
        self._menu_button.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)
        self._menu_button.clicked.connect(self._on_menu_callback)

        widget_layout.addWidget(self._widget)
        widget_layout.addWidget(self._menu_button)

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        self._error_label = QLabel(self)
        self._error_label.setHidden(True)
        self._error_label.setObjectName('errorLabel')
        self._error_label.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)

        layout.addLayout(widget_layout)
        layout.addWidget(self._error_label)

        self._layout2.addLayout(layout)

    def is_required(self):
        """
        Returns whether current field is required for the field widget or not
        :return: bool
        """

        return bool(self._required)

    def set_required(self, flag):
        """
        Sets whether the field is required for the field widget or not
        :param flag: bool
        """

        self._required = flag
        self.setProperty('required', flag)
        self.setStyleSheet(self.styleSheet())

    def reset(self):
        """
        Resets the field widget back to its default values
        """

        self.set_state(self._data)

    def refresh(self):
        """
        Refresh the style properties of the field
        """

        direction = self._data.get('layout', self.DefaultLayout)
        self.setProperty('layout', direction)
        self.setProperty('default', self.is_default())
        self.setProperty('error', self._error)
        self.setStyleSheet(self.styleSheet())

    def _action_callback(self, callback):
        """
        Internal function that wraps schema callback to get the return value
        :param callback: fn
        """

        self._action_result = callback()

    def _on_menu_callback(self):
        """
        Internal callback function that is triggered when the menu button is clicked
        """

        callback = self.data().get('menu', {}).get('callback', self._on_how_menu)
        callback()

    def _on_show_menu(self):
        """
        Internal callback that shows field menu using the actions from the data
        """

        menu = QMenu(self)

        actions = self.data().get('actions', [])
        for action in actions:
            name = action.get('name', 'No name found')
            callback = action.get('callback')
            fn = partial(self._action_callback, callback)
            action = menu.addAction(name)
            action.triggered.connect(fn)

        point = QCursor.pos()
        point.setX(point.x() + 3)
        point.setY(point.y() + 3)

        self._action_result = None

        menu.exec_(point)

        if self._action_result is not None:
            self.set_value(self._action_result)

    def _on_emit_value_changed(self, *args):
        """
        Emits the value changed signal
        :param args: list
        """

        self.valueChanged.emit()
        self.refresh()


class BoolFieldWidget(FieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(BoolFieldWidget, self).__init__(*args, **kwargs)

        widget = QCheckBox(self)
        widget.stateChanged.connect(self._on_emit_value_changed)
        self.set_widget(widget)
        inline = self.data().get('inline')
        if inline:
            self.label().setText('')
            self.widget().setText(self.title())

    def value(self):
        """
        Implements FieldWidget value function
        Returns the value of the checkbox
        :return: str
        """

        return bool(self.widget().isChecked())

    def set_value(self, value):
        """
        Overrides FileWidget set_value function
        Sets the value of the checkbox
        :param value: str
        """

        self.widget().setChecked(value)
        super(BoolFieldWidget, self).set_value(value)


class IntFieldWidget(FieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(IntFieldWidget, self).__init__(*args, **kwargs)

        validator = QIntValidator(-sys.maxint, sys.maxint, self)

        widget = QLineEdit(self)
        widget.setValidator(validator)
        widget.stateChanged.connect(self._on_emit_value_changed)
        self.set_widget(widget)

    def value(self):
        """
        Implements FieldWidget value function
        Returns the value of the widget
        :return: str
        """

        value = self.widget().text()
        if value.strip() == '':
            value = self.default()

        return int(str(value))

    def set_value(self, value):
        """
        Overrides FileWidget set_value function
        Sets the value of the widget
        :param value: str
        """

        if value == '':
            value = self.default()
        self.widget().setText(str(int(value)))


class SliderFieldWidget(FieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(SliderFieldWidget, self).__init__(*args, **kwargs)

        widget = QSlider(self)
        widget.setOrientation(Qt.Horizontal)
        widget.setObjectName('widget')
        widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        widget.valueChanged.connect(self._on_emit_value_changed)
        self.set_widget(widget)

    def value(self):
        """
        Implements FieldWidget value function
        Returns the value of the slider
        :return: str
        """

        return self.widget().value()

    def set_value(self, value):
        """
        Overrides FileWidget set_value function
        Sets the value of the slider
        :param value: str
        """

        self.widget().setValue(value)


class RangeFieldWidget(FieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(RangeFieldWidget, self).__init__(*args, **kwargs)

        widget = QFrame(self)
        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(3)
        widget.setLayout(layout)

        validator = QIntValidator(-sys.maxint, sys.maxint, self)

        self._min_widget = QLineEdit(self)
        self._min_widget.setValidator(validator)
        self._min_widget.textChanged.connect(self._on_emit_value_changed)
        widget.layout().addWidget(self._min_widget)

        self._max_widget = QLineEdit(self)
        self._max_widget.setValidator(validator)
        self._max_widget.textChanged.connect(self._on_emit_value_changed)
        widget.layout().addWidget(self._max_widget)

        self.set_widget(widget)

    def value(self):
        """
        Implements FieldWidget value function
        Returns the current range
        :return: list(int)
        """

        min_value = int(float(self._min_widget.text() or '0'))
        max_value = int(float(self._max_widget.text() or '0'))

        return min_value, max_value

    def set_value(self, value):
        """
        Overrides FileWidget set_value function
        Sets the current range
        :param value: list(int)
        """

        min_value, max_value = int(value[0], int(value[1]))
        self._min_widget.setText(str(min_value))
        self._max_widget.setText(str(max_value))

        super(RangeFieldWidget, self).set_value(value)


class StringFieldWidget(FieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(StringFieldWidget, self).__init__(*args, **kwargs)

        widget = label.QLineEdit(self)
        widget.textChanged.connect(self._on_emit_value_changed)
        self.set_widget(widget)

    def value(self):
        """
        Implements FieldWidget value function
        Returns the value of the widget
        :return: str
        """

        return str(self.widget().text())

    def set_value(self, value):
        """
        Overrides FileWidget set_value function
        Sets the value of the widget
        :param value: str
        """

        self.widget().setText(value)
        super(StringFieldWidget, self).set_value(value)


class PasswordFieldWidget(FieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(PasswordFieldWidget, self).__init__(*args, **kwargs)

        widget = label.QLineEdit(self)
        widget.setEchoMode(QLineEdit.EchoMode.Password)
        widget.textChanged.connect(self._on_emit_value_changed)
        self.set_widget(widget)

    def value(self):
        """
        Implements FieldWidget value function
        Returns the value of the widget
        :return: str
        """

        return str(self.widget().text())

    def set_value(self, value):
        """
        Overrides FileWidget set_value function
        Sets the value of the widget
        :param value: str
        """

        self.widget().setText(value)
        super(PasswordFieldWidget, self).set_value(value)


class TextFieldWidget(FieldWidget, object):

    DefaultLayout = 'vertical'

    def __init__(self, *args, **kwargs):
        super(TextFieldWidget, self).__init__(*args, **kwargs)

        widget = label.QTextEdit(self)
        widget.textChanged.connect(self._on_emit_value_changed)
        self.set_widget(widget)

    def value(self):
        """
        Implements FieldWidget value function
        Returns the value of the text edit
        :return: str
        """

        return str(self.widget().toPlainText())

    def set_value(self, value):
        """
        Overrides FileWidget set_value function
        Sets the value of the text edit
        :param value: str
        """

        self.widget().setText(value)
        super(TextFieldWidget, self).set_value(value)


class PathFieldWidget(StringFieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(PathFieldWidget, self).__init__(*args, **kwargs)

    def set_data(self, data):
        """
        Overrides StringFieldWidget set_data function
        Adds a browse button to folder button
        :param data: dict
        """

        if 'menu' not in data:
            data['menu'] = {
                'callback': self._on_browse
            }
        super(PathFieldWidget, self).set_data(data)

    def _on_browse(self):
        """
        Opens file dialog
        """

        path = self.value()
        path = QFileDialog.getExistingDirectory(None, 'Browse Folder', path)
        if path:
            self.set_value(path)

        widget = QCheckBox(self)
        widget.stateChanged.connect(self._on_emit_value_changed)
        self.set_widget(widget)
        inline = self.data().get('inline')
        if inline:
            self.label().setText('')
            self.widget().setText(self.title())


class LabelFieldWidget(FieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(LabelFieldWidget, self).__init__(*args, **kwargs)

        widget = label.RightElidedLabel(self)
        widget.setAlignment(Qt.AlignVCenter)
        widget.setTextInteractionFlags(Qt.TextSelectableByMouse)
        self.set_widget(widget)

    def value(self):
        """
        Implements FieldWidget value function
        Returns the value of the label
        :return: str
        """

        return str(self.widget().text())

    def set_value(self, value):
        """
        Overrides FileWidget set_value function
        Sets the value of the label
        :param value: str
        """

        self.widget().setText(value)
        super(LabelFieldWidget, self).set_value(value)


class EnumFieldWidget(FieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(EnumFieldWidget, self).__init__(*args, **kwargs)

        widget = QComboBox(self)
        widget.currentIndexChanged.connect(self._on_emit_value_changed)
        self.set_widget(widget)

    def value(self):
        """
        Implements FieldWidget value function
        Returns the value of the combo box
        :return: str
        """

        return str(self.widget().currentText())

    def set_value(self, item):
        """
        Overrides FileWidget set_value function
        Sets the value of the combo box
        :param item: str
        """

        self.set_current_text(item)

    def set_items(self, items):
        """
        Overrides FieldWidget set_items function
        Sets the current items of the combo box
        :param items: list(str)
        """

        self.widget().clear()
        self.widget().addItems(items)

    def set_state(self, state):
        """
        Sets the current state with support for editable
        :param state: dict
        """

        super(EnumFieldWidget, self).set_state(state)
        editable = state.get('editable')
        if editable is not None:
            self.widget().setEditable(editable)

    def set_current_text(self, text):
        """
        Sets current text
        :param text: str
        """

        index = self.widget().findText(text, Qt.MatchExactly)
        if index != -1:
            self.widget().setCurrentIndex(index)
        else:
            qt.logger.warning('Cannot set the value for field {}'.format(self.name()))


class SeparatorFieldWidget(FieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(SeparatorFieldWidget, self).__init__(*args, **kwargs)

        widget = QLabel(self)
        widget.setObjectName('widget')
        widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        self.label().hide()
        self.set_widget(widget)

    def value(self):
        """
        Implements FieldWidget value function
        Returns the value of the separator
        :return: str
        """

        return str(self.widget().text())

    def set_value(self, value):
        """
        Overrides FileWidget set_value function
        Sets the value of the separator
        :param value: str
        """

        self.widget().setText(value)


class ColorFieldWidget(FieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(ColorFieldWidget, self).__init__(*args, **kwargs)

        widget = color.ColorPicker()
        widget.setObjectName('widget')
        widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        widget.colorChanged.connect(self._on_color_changed)
        self.set_widget(widget)

    def set_data(self, data):
        """
        Overrides FieldWidget data function
        :param data: dict
        """

        colors = data.get('colors')
        if colors:
            self.widget().set_colors(colors)

        super(ColorFieldWidget, self).set_data(data)

    def value(self):
        """
        Implements FieldWidget value function
        Returns the value of the color picker
        :return: Color
        """

        return self.widget().current_color()

    def set_value(self, value):
        """
        Overrides FileWidget set_value function
        Sets the value of the color picker
        :param value: str
        """

        self.widget().set_current_color(value)

    def _on_color_changed(self, new_color):
        """
        Internal callback function triggered when the color changes from the color browser
        :param new_color: QColor
        """

        self.set_value(new_color)
        self._on_emit_value_changed()


class ImageFieldWidget(FieldWidget, object):
    def __init__(self, *args, **kwargs):
        super(ImageFieldWidget, self).__init__(*args, **kwargs)

        self._value = ''
        self._pixmap = None

        widget = QLabel(self)
        widget.setObjectName('widget')
        self.setStyleSheet('min-height: 32px;')
        widget.setScaledContents(False)
        widget.setAlignment(Qt.AlignHCenter)
        self.set_widget(widget)
        self.layout().addStretch()

    def resizeEvent(self, event):
        """
        Overrides FieldWidget resizeEvent function
        Called when teh field widget is resized
        :param event: QResizeEvent
        """

        self.update()

    def value(self):
        """
        Implements FieldWidget value function
        Returns the path of the image in disk
        :return: str
        """

        return self._value

    def set_value(self, value):
        """
        Overrides FileWidget set_value function
        Sets the path of the image in disk
        :param value: str
        """

        self._value = value
        self._pixmap = QPixmap(value)
        self.update()

    def update(self):
        """
        Updates the image depending on the size
        """

        if not self._pixmap:
            return

        width = self.widget().height()
        if self.widget().width() > self.widget().height():
            pixmap = self._pixmap.scaledToWidth(width, Qt.SmoothTransformation)
        else:
            pixmap = self._pixmap.scaledToHeight(width, Qt.SmoothTransformation)
        self.widget().setPixmap(pixmap)
        self.widget().setAlignment(Qt.AlignLeft)


FIELD_WIDGET_REGISTRY = {
    "int": IntFieldWidget,
    "bool": BoolFieldWidget,
    "enum": EnumFieldWidget,
    "text": TextFieldWidget,
    "path": PathFieldWidget,
    "image": ImageFieldWidget,
    "label": LabelFieldWidget,
    "range": RangeFieldWidget,
    "color": ColorFieldWidget,
    "string": StringFieldWidget,
    "password": PasswordFieldWidget,
    "slider": SliderFieldWidget,
    "separator": SeparatorFieldWidget
}



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains different graphic views
"""

from __future__ import print_function, division, absolute_import

import os
import math

from Qt.QtCore import *
from Qt.QtWidgets import *
from Qt.QtGui import *

from tpDcc.libs.python import mathlib

try:
    from Qt import QtOpenGL
except ImportError:
    try:
        from PySide import QtOpenGL
    except ImportError:
        # Max 2018 does not support QtOpenGL?
        tp.logger.warning('QtOpenGL is not available! QtOpenGL functionality will not be available!')


class ViewportModes(object):
    Full = 'full'
    Smart = 'smart'
    Minimal = 'minimal'
    Bounding = 'bounding'


class AutoPanController(object):
    def __init__(self, amount=10.0):
        super(AutoPanController, self).__init__()

        self._enabled = False
        self._amount = amount
        self._auto_pan_delta = QVector2D(0.0, 0.0)
        self._been_outside = False

    def get_amount(self):
        return self._amount

    def set_amount(self, amount):
        self._amount = amount

    def get_auto_pan_delta(self):
        return self._auto_pan_delta

    def get_enabled(self):
        return self._enabled

    amount = property(get_amount, set_amount)
    auto_pan_delta = property(get_auto_pan_delta)
    enabled = property(get_enabled)

    def start(self):
        self._enabled = True

    def update(self, rect, pos):
        if self._enabled:
            if pos.x() < 0:
                self._auto_pan_delta = QVector2D(-self._amount, 0.0)
                self._been_outside = True
                self._amount = mathlib.clamp(abs(pos.x()) * 0.3, 0.0, 25.0)
            if pos.x() > rect.width():
                self._auto_pan_delta = QVector2D(self._amount, 0.0)
                self._been_outside = True
                self._amount = mathlib.clamp(abs(rect.width() - pos.x()) * 0.3, 0.0, 25.0)
            if pos.y() < 0:
                self._auto_pan_delta = QVector2D(0.0, -self._amount)
                self._been_outside = True
                self._amount = mathlib.clamp(abs(pos.y()) * 0.3, 0.0, 25.0)
            if pos.y() > rect.height():
                self._auto_pan_delta = QVector2D(0.0, self._amount)
                self._been_outside = True
                self._amount = mathlib.clamp(abs(rect.height() - pos.y()) * 0.3, 0.0, 25.0)
            if self._been_outside and rect.contains(pos):
                self.reset()

    def stop(self):
        self._enabled = False
        self.reset()

    def reset(self):
        self._been_outside = False
        self._auto_pan_delta = QVector2D(0.0, 0.0)


class RubberRect(QGraphicsRectItem, object):

    DEFAULT_RUBBER_RECT_COLOR = QColor(255, 255, 255, 50)

    def __init__(self, name):
        super(RubberRect, self).__init__()
        self._name = name
        self.setZValue(2)
        self.setPen(QPen(self.DEFAULT_RUBBER_RECT_COLOR, 0.5, Qt.SolidLine))
        self.setBrush(QBrush(self.DEFAULT_RUBBER_RECT_COLOR))

    def get_name(self):
        return self._name

    def set_name(self, name):
        self._name = name

    name = property(get_name, set_name)


class BaseGraphicsView(QGraphicsView, object):
    """
    QGraphicsView with custom functionality:
        - Zoom in/Zoom out
        - Panning/Auto Panning
        - Custom Rubber Rect Selection
    """

    selectionChanged = Signal()

    def __init__(self, min_scale=0.5, max_scale=2.0, parent=None, **kwargs):
        super(BaseGraphicsView, self).__init__(parent=parent)

        self._min_scale = min_scale
        self._max_scale = max_scale
        self._factor = 1.0
        self._factor_diff = 0.0
        self._scale = 1.0
        self._pan_speed = 1.0
        self._is_panning = False
        self._init_scrollbars_pos = QVector2D(self.horizontalScrollBar().value(), self.verticalScrollBar().value())
        self._is_rubber_rect_selection = False
        self._use_opengl = kwargs.get('use_opengl', False)
        self.log = kwargs.get('log', tp.logger)

        self._mouse_pressed = False
        self._pressed_item = None
        self._released_item = None
        self._mouse_pressed_pos = QPoint(0, 0)
        self._mouse_release_pos = QPoint(0, 0)
        self._mouse_pos = QPoint(0, 0)
        self._last_mouse_pos = QPoint(0, 0)

        self._auto_pan_controller = AutoPanController()
        self._rubber_rect = RubberRect(name='RubberRect')

        self.initialize_scene_view()

        self.setInteractive(True)
        self.setMouseTracking(True)
        self.setRenderHint(QPainter.Antialiasing, True)
        self.setRenderHint(QPainter.TextAntialiasing, True)
        self.setRenderHint(QPainter.HighQualityAntialiasing, True)
        self.setRenderHint(QPainter.SmoothPixmapTransform, True)
        self.setRenderHint(QPainter.NonCosmeticDefaultPen, True)
        self.setAttribute(Qt.WA_AlwaysShowToolTips)
        self.setRubberBandSelectionMode(Qt.IntersectsItemShape)

    # region Properties
    def get_viewport_mode(self):
        mode = self.viewportUpdateMode()
        if mode == QGraphicsView.FullViewportUpdate:
            return ViewportModes.Full
        if mode == QGraphicsView.SmartViewportUpdate:
            return ViewportModes.Smart
        if mode == QGraphicsView.MinimalViewport:
            return ViewportModes.Minimal
        if mode == QGraphicsView.BoundingRectViewportUpdate:
            return ViewportModes.Bounding

        return None

    def set_viewport_mode(self, mode):
        if mode == ViewportModes.Full:
            mode = QGraphicsView.FullViewportUpdate
        if mode == ViewportModes.Smart:
            mode = QGraphicsView.SmartViewportUpdate
        if mode == ViewportModes.Minimal:
            mode = QGraphicsView.MinimalViewportUpdate
        if mode == ViewportModes.Bounding:
            mode = QGraphicsView.BoundingRectViewportUpdate

        self.setViewportUpdateMode(mode)
    # endregion

    # region Override Functions
    def mousePressEvent(self, event):
        super(BaseGraphicsView, self).mousePressEvent(event)

        self._pressed_item = self.itemAt(event.pos())
        self._mouse_pressed_pos = event.pos()

        modifiers = event.modifiers()

        if self._pressed_item and isinstance(self._pressed_item, QGraphicsItem):
            self._auto_pan_controller.start()

        if not self._pressed_item:
            if event.button() == Qt.LeftButton:
                self._is_rubber_rect_selection = True
            if event.button() == Qt.RightButton and modifiers == Qt.NoModifier:
                self._is_panning = True
            self._init_scrollbars_pos = QVector2D(self.horizontalScrollBar().value(), self.verticalScrollBar().value())

    def mouseMoveEvent(self, event):
        super(BaseGraphicsView, self).mouseMoveEvent(event)

        self._mouse_pos = event.pos()

        if self._is_panning:
            delta = self.mapToScene(event.pos()) - self.mapToScene(self._last_mouse_pos)
            self._pan(delta)

        if self._rubber_rect and self.scene():
            if self._is_rubber_rect_selection:
                current_pos = self.mapToScene(self._mouse_pos)
                press_pos = self.mapToScene(self._mouse_pressed_pos)
                if self._rubber_rect not in self.scene().items():
                    self.scene().addItem(self._rubber_rect)
                if not self._rubber_rect.isVisible():
                    self._rubber_rect.setVisible(True)
                r = QRectF(
                    press_pos.x(), press_pos.y(), current_pos.x() - press_pos.x(), current_pos.y() - press_pos.y())
                self._rubber_rect.setRect(r.normalized())

        self._auto_pan_controller.update(self.viewport().rect(), event.pos())

        self._last_mouse_pos = event.pos()

    def mouseReleaseEvent(self, event):
        super(BaseGraphicsView, self).mouseReleaseEvent(event)

        self._auto_pan_controller.stop()
        self._mouse_release_pos = event.pos()
        self._released_item = self.itemAt(event.pos())
        self._is_panning = False

        if self._rubber_rect and self.scene():
            if self._is_rubber_rect_selection:
                self._is_rubber_rect_selection = False
                self.setDragMode(QGraphicsView.NoDrag)
                self._select_rubber_rect_items()
                self.remove_item_by_name(self._rubber_rect.name)

    def keyPressEvent(self, event):
        modifiers = event.modifiers()

        # Zoom In
        if all([event.key() == Qt.Key_Equal, modifiers == Qt.ControlModifier]):
            # TODO: Check why is not working
            self.zoom_delta(True)

        # Zoom Out
        if all([event.key() == Qt.Key_Minus, modifiers == Qt.ControlModifier]):
            self.zoom_delta(False)

        # Rest Graph Zoom
        if all([event.key() == Qt.Key_R, modifiers == Qt.ControlModifier]):
            self.reset_scale()

        super(BaseGraphicsView, self).keyPressEvent(event)

    def wheelEvent(self, event):
        self._zoom(math.pow(2.0, event.delta() / 240.0))
    # endregion

    # region Public Functions
    def initialize_scene_view(self):
        """
        Instantiates the proper Scene and enables OpenGL if necessary
        Scene instantiation must be override in child classes
        """

        if self._use_opengl:
            self.setViewport(QtOpenGL.QGLWidget(QtOpenGL.QGLFormat(QtOpenGL.QGL.SampleBuffers)))
            self.log.info('Initializing OpenGL renderer')

    def get_contents_size(self):
        """
        Returns the view contents size (pyshical size)
        :return: tuple, content size
        """

        content_rect = self.contentsRect()
        return (content_rect.width(), content_rect.height())

    def get_center(self):
        """
        Returns the correct center point of the current view
        :return: QPointF, current view center point
        """

        return self.mapToScene(self.viewport().rect().center())

    def set_center(self, pos):
        """
        Sets the current scene center point
        :param pos: tuple, x and y coordinates
        """

        self.centerOn(pos[0], pos[1])

    def get_scene_coordinates(self):
        """
        Returns the scene size
        :return: tuple, coordinates of current scene (-x, -y, x, y)
        """

        if self.scene():
            return self.scene().sceneRect().getCoords()
        return (0, 0, 0, 0)

    def get_translation(self):
        """
        Returns the current scrollbar positions
        :return: tuple, scroll bar coordinates (h, v)
        """

        return [self.horizontalScrollBar().value(), self.verticalScrollBar().value()]

    def get_scale_factor(self):
        """
        Returns the current scale factor
        """

        return [self.transform().m11(), self.transform().m22()]

    def get_scene_attributes(self):
        """
        Returns a dictionary of scene attributes
        :return: dict, view position, scale, size
        """

        scene_attributes = dict()
        scene_attributes.update(view_scale=self.get_scale_factor())
        scene_attributes.update(view_center=self.get_center())
        scene_attributes.update(view_size=self.get_contents_size())
        scene_attributes.update(scene_size=self.get_scene_coordinates())

        return scene_attributes

    def update_scene_attributes(self):
        """
        Updates scene attributes
        """

        scene_attributes = self.get_scene_attributes()
        return scene_attributes

    def remove_item_by_name(self, name):
        """
        Remove graph item by its name (can be any kind of QGraphicsItem)
        :param name: str, name of the item to delete
        """

        if self.scene():
            [self.scene().removeItem(i) for i in self.scene().items() if hasattr(i, 'name') and i.name == name]

    def move_scrollbar(self, delta):
        """
        Move the scrollbar of the view by a given delta
        :param delta: QPointF
        """

        x = self.horizontalScrollBar().value() + delta.x()
        y = self.verticalScrollBar().value() + delta.y()
        self.horizontalScrollBar().setValue(x)
        self.verticalScrollBar().setValue(y)

    def set_scrollbars_positions(self, horizontal, vertical):
        """
        Set the position of the view scrollbars
        :param horizontal: float, position of the horizontal scrollbar
        :param vertical: float, position of the vertical scrollbar
        :return:
        """

        try:
            self.horizontalScrollBar().setValue(horizontal)
            self.verticalScrollBar().setValue(vertical)
        except Exception as e:
            tp.logger.debug(str(e))

    def reset_scale(self):
        self.resetMatrix()

    def zoom_delta(self, direction):
        if direction:
            self._zoom(1 + 0.1)
        else:
            self._zoom(1 - 0.1)

    def fit_scene_content(self):
        """
        Fits scene content to view, by scaling it
        """

        if self.scene():
            scene_rect = self.scene().get_bounding_rect(margin=0)
            self.fitInView(scene_rect, Qt.KeepAspectRatio)
    # endregion

    # region Private Functions
    def _pan(self, delta):
        delta *= self._scale * -1
        delta *= self._pan_speed
        self.horizontalScrollBar().setValue(self.horizontalScrollBar().value() + delta.x())
        self.verticalScrollBar().setValue(self.verticalScrollBar().value() + delta.y())

    def _zoom(self, scale_factor):
        self._factor = round(self.matrix().scale(scale_factor, scale_factor).mapRect(QRectF(0, 0, 1, 1)).width(), 1)

        if hasattr(self.scene(), 'grid_size'):
            if self._factor < (self._min_scale + 0.4):
                self.scene().grid_size = 20
            else:
                self.scene().grid_size = 10

        if self._factor < self._min_scale or self._factor > self._max_scale:
            return
        self.scale(scale_factor, scale_factor)
        self._scale *= scale_factor

    def _select_rubber_rect_items(self):
        self.scene().blockSignals(True)
        items = [i for i in self._rubber_rect.collidingItems()]
        for item in items[:-1]:
            item.setSelected(True)
        self.scene().blockSignals(False)
        if len(items) > 0:
            items[-1].setSelected(True)


class GridView(BaseGraphicsView, object):
    """
    View with grid drawing support
    """

    def __init__(self, parent=None):
        super(GridView, self).__init__(parent=parent)

        self._grid_spacing = 100
        self._grid_size = 10
        self._draw_grid_size = self._grid_size * 2
        self._show_grid = True

        self.setRenderHint(QPainter.Antialiasing)

        self.setViewportUpdateMode(QGraphicsView.BoundingRectViewportUpdate)
        self.setCacheMode(QGraphicsView.CacheBackground)
        self.setAttribute(Qt.WA_AlwaysShowToolTips)

    def drawBackground(self, painter, rect):
        super(GridView, self).drawBackground(painter, rect)

        if self._show_grid:
            scene_rect = self.sceneRect()
            left = int(scene_rect.left()) - (int(scene_rect.left()) % self._draw_grid_size)
            top = int(scene_rect.top()) - (int(scene_rect.top()) % self._draw_grid_size)
            scale_mult = 1.0

            lines = list()
            left_line = rect.left() - rect.left() % self._grid_size
            top_line = rect.top() - rect.top() % self._grid_size
            i = int(left_line)
            while i < int(rect.right()):
                lines.append(QLineF(i, rect.top(), i, rect.bottom()))
                i += self._grid_size
            u = int(top_line)
            while u < int(rect.bottom()):
                lines.append(QLineF(rect.left(), u, rect.right(), u))
                u += self._grid_size
            # TODO: Change pen to a class variable (avoid to create a pen each drawing frame)
            pen = QPen()
            pen.setWidth(0)
            pen.setColor(QColor(20, 20, 20))
            painter.setPen(pen)
            painter.drawLines(lines)


class GridBackgroundImageView(GridView, object):
    """
    View with image background drawing support
    """

    def __init__(self, parent=None):
        super(GridBackgroundImageView, self).__init__(parent=parent)

        self._background_image = None
        self._background_image_path = None
        self._fit_image_to_window = False

        self.setBackgroundBrush(QBrush(QColor(70, 70, 70, 255)))

    def get_background_image(self):
        return self._background_image

    def set_background_image(self, value, fit_to_window=False, mirror_x=False, mirror_y=False):
        if not value:
            return
        value = str(value)
        if not (value and os.path.exists(value)):
            print('background image not found: {}'.format(value))
            return

        self._background_image_path = value
        self._fit_image_to_window = fit_to_window

        # Load image and mirror it vertically
        self._background_image = QImage(value).mirrored(mirror_x, mirror_y)

        if not fit_to_window:
            width = self._background_image.width()
            height = self._background_image.height()
            self.scene().set_size(width, height)
            self.fit_scene_content()

    background_image = property(get_background_image, set_background_image)

    # def resizeEvent(self, event):
    # TODO: tpoveda: This is not working when an image is not loaded
    #     if self._fit_image_to_window:
    #         self.scene().set_size(self.rect().width(), self.rect().height())
    #     else:
    #         self.fit_scene_content()
    #     super(GridBackgroundImageView, self).resizeEvent(event)

    def drawBackground(self, painter, rect):
        """
        Override method to draw view custom background image
        """

        if not self.background_image:
            return super(GridBackgroundImageView, self).drawBackground(painter, rect)
        super(GridBackgroundImageView, self).drawBackground(painter, rect)
        painter.drawImage(self.sceneRect(), self.background_image, QRectF(self.background_image.rect()))



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains base class to handle DPI functionality
"""

from __future__ import print_function, division, absolute_import

from tpDcc.libs.qt.core import qtutils


class DPIScaling(object):
    """
    Mixin class that can be used in any QWidget to add DPI scaling functionality to it
    """

    def setFixedSize(self, size):
        return super(DPIScaling, self).setFixedSize(qtutils.dpi_scale(size))

    def setFixedHeight(self, height):
        return super(DPIScaling, self).setFixedHeight(qtutils.dpi_scale(height))

    def setFixedWidth(self, width):
        return super(DPIScaling, self).setFixedWidth(qtutils.dpiScale(width))

    def setMaximumWidth(self, width):
        return super(DPIScaling, self).setMaximumWidth(qtutils.dpi_scale(width))

    def setMinimumWidth(self, width):
        return super(DPIScaling, self).setMinimumWidth(qtutils.dpi_scale(width))

    def setMaximumHeight(self, height):
        return super(DPIScaling, self).setMaximumHeight(qtutils.dpi_scale(height))

    def setMinimumHeight(self, height):
        return super(DPIScaling, self).setMinimumHeight(qtutils.dpi_scale(height))



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains extended functionality for QMenus
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import *
from Qt.QtWidgets import *

from tpDcc.libs.python import python
from tpDcc.libs.qt.core import mixin, qtutils, formatters


@mixin.theme_mixin
@mixin.property_mixin
class BaseMenu(QMenu, object):

    valueChanged = Signal(list)

    def __init__(self, exclusive=True, cascader=False, title='', parent=None):
        super(BaseMenu, self).__init__(title=title, parent=parent)

        self._load_data_fn = None

        self._action_group = QActionGroup(self)
        self._action_group.setExclusive(exclusive)
        self._action_group.triggered.connect(self._on_action_triggered)

        self.setProperty('cascader', cascader)
        self.setCursor(Qt.PointingHandCursor)

        self.set_value('')
        self.set_data([])
        self.set_separator('/')

    # =================================================================================================================
    # BASE
    # =================================================================================================================

    def set_separator(self, separator_character):
        """
        Sets menu separator character
        :param separator_character: str
        """
        self.setProperty('separator', separator_character)

    def set_value(self, data):
        """
        Sets menu value
        :param data:  str, int or float
        """

        assert isinstance(data, (list, str, unicode, int, float))
        if self.property('cascader') and isinstance(data, (str, unicode)):
            data = data.split(self.property('separator'))
        self.setProperty('value', data)

    def set_data(self, option_list):
        """
        Sets menu data
        :param option_list: list
        """

        assert isinstance(option_list, list)
        if option_list:
            if all(isinstance(i, (str, unicode)) for i in option_list):
                option_list = python.from_list_to_nested_dict(option_list, separator=self.property('separator'))
            if all(isinstance(i, (int, float)) for i in option_list):
                option_list = [{'value': i, 'label': str(i)} for i in option_list]
        self.setProperty('data', option_list)

    def set_loader(self, fn):
        """
        Sets menu loader
        :param fn: function
        """

        self._load_data_fn = fn

    def set_load_callback(self, fn):
        """
        Sets menu load callback
        :param fn: function
        """

        assert callable(fn)
        self._load_data_fn = fn
        self.aboutToShow.connect(self._on_fetch_data)

    # =================================================================================================================
    # PROPERTY MIXIN SETTERS
    # =================================================================================================================

    def _set_value(self, value):
        data_list = value if isinstance(value, list) else [value]
        flag = False
        for act in self._action_group.actions():
            checked = act.property('value') in data_list
            if act.isChecked() != checked:
                act.setChecked(checked)
                flag = True

        if flag:
            self.valueChanged.emit(value)

    def _set_data(self, option_list):
        self.clear()
        for act in self._action_group.actions():
            self._action_group.removeAction(act)
        for data_dict in option_list:
            self._add_menu(self, data_dict)

    # =================================================================================================================
    # INTERNAL
    # =================================================================================================================

    def _add_menu(self, parent_menu, data_dict):
        if 'children' in data_dict:
            menu = BaseMenu(title=data_dict.get('label'), parent=self)
            menu.setProperty('value', data_dict.get('value'))
            parent_menu.addMenu(menu)
            if parent_menu is not self:
                menu.setProperty('parent_menu', parent_menu)
            for i in data_dict.get('children'):
                self._add_menu(menu, i)
        else:
            action = self._action_group.addAction(formatters.display_formatter(data_dict.get('label')))
            action.setProperty('value', data_dict.get('value'))
            action.setCheckable(True)
            action.setProperty('parent_menu', parent_menu)
            parent_menu.addAction(action)

    def _get_parent(self, result, obj):
        if obj.property('parent_menu'):
            parent_menu = obj.property('parent_menu')
            result.insert(0, parent_menu.title())
            self._get_parent(result, parent_menu)

    # =================================================================================================================
    # CALLBACKS
    # =================================================================================================================

    def _on_action_triggered(self, action):
        current_data = action.property('value')
        if self.property('cascader'):
            selected_data = [current_data]
            self._get_parent(selected_data, action)
        else:
            if self._action_group.isExclusive():
                selected_data = current_data
            else:
                selected_data = [act.property('value') for act in self._action_group.actions() if act.isChecked()]
        self.set_value(selected_data)
        self.valueChanged.emit(selected_data)

    def _on_fetch_data(self):
        data_list = self._load_data_fn()
        self.set_data(data_list)


class Menu(QMenu, object):

    mouseButtonClicked = Signal(object, object)     # mouseButton, QAction

    def __init__(self, *args, **kwargs):
        super(Menu, self).__init__(*args, **kwargs)

    def mouseReleaseEvent(self, event):
        """
        Extends mouseReleaseEvent QMenu function
        :param event: QMouseEvent
        """

        self.mouseButtonClicked.emit(event.button(), self.actionAt(event.pos()))

        return super(Menu, self).mouseReleaseEvent(event)

    def insertAction(self, before, *args):
        """
        Extends insertAction QMenu function
        Add supports for finding the before action by the given string
        :param before: str or QAction
        :param args: list
        :return: QAction
        """

        if isinstance(before, (unicode, str)):
            before = self.find_action(before)

        return super(Menu, self).insertAction(before, *args)

    def insertMenu(self, before, menu):
        """
        Extends insertMenu QMenu function
        Add supports for finding the before action by the given string
        :param before: str or QAction
        :param menu: QMenu
        :return: QAction
        """

        if isinstance(before, (unicode, str)):
            before = self.find_action(before)

        return super(Menu, self).insertMenu(before, menu)

    def insertSeparator(self, before):
        """
        Extends insertSeparator QMenu function
        :param before: str or QAction
        :return: QAction
        """

        if isinstance(before, (unicode, str)):
            before = self.find_action(before)

        return super(Menu, self).insertSeparator(before)

    def find_action(self, text):
        """
        Returns the action that contains the given text
        :param text: str
        :return: QAction
        """

        for child in self.children():
            action = None
            if isinstance(child, QMenu):
                action = child.menuAction()
            elif isinstance(child, QAction):
                action = child
            if action and action.text().lower() == text.lower():
                return action


class SearchableTaggedAction(QAction, object):
    def __init__(self, label, icon=None, parent=None):
        super(SearchableTaggedAction, self).__init__(label, parent)

        self._tags = set()

        if icon:
            self.setIcon(icon)

    @property
    def tags(self):
        return self._tags

    @tags.setter
    def tags(self, new_tags):
        self._tags = new_tags

    def has_tag(self, tag):
        """
        Searches this instance tags. Returns True if the tag is valid or False otherwise
        :param str tag: partial or full tag to search for
        :return: bool
        """

        for t in self._tags:
            if tag in t:
                return True

        return False

    def has_any_tag(self, tags):
        """
        Returns True if current action has some of the given tags or False otherwise
        :param tags: list(str)
        :return: bool
        """

        for t in tags:
            for i in self._tags:
                if t in i:
                    return True

        return False


class SearchableMenu(Menu, object):
    """
    Extends standard QMenu to make it searchable. First action is a QLineEdit used to recursively search on all actions
    """

    def __init__(self, **kwargs):
        super(SearchableMenu, self).__init__(**kwargs)

        self._search_action = None
        self._search_edit = None

        self.setObjectName(kwargs.get('objectName'))
        self.setTitle(kwargs.get('title'))
        self._init_search_edit()

    def clear(self):
        """
        Extends QMenu clear function
        """

        super(SearchableMenu, self).clear()

        self._init_search_edit()

    def set_search_visible(self, flag):
        """
        Sets the visibility of the search edit
        :param flag: bool
        """

        self._search_action.setVisible(flag)

    def search_visible(self):
        """
        Returns whether or not search edit is visible
        :return: bool
        """

        return self._search_action.isVisible()

    def update_search(self, search_string=None):
        """
        Search all actions for a string tag
        :param str search_string: tag names separated by spaces (for example, "elem1 elem2"
        :return: str
        """

        def _recursive_search(menu, search_str):
            for action in menu.actions():
                sub_menu = action.menu()
                if sub_menu:
                    _recursive_search(sub_menu, search_str)
                    continue
                elif action.isSeparator():
                    continue
                elif isinstance(action, SearchableTaggedAction) and not action.has_tag(search_str):
                    action.setVisible(False)

            menu_vis = any(action.isVisible() for action in menu.actions())
            menu.menuAction().setVisible(menu_vis)

        def _recursive_search_by_tags(menu, tags):
            for action in menu.actions():
                sub_menu = action.menu()
                if sub_menu:
                    _recursive_search_by_tags(sub_menu, tags)
                    continue
                elif action.isSeparator():
                    continue
                elif isinstance(action, SearchableTaggedAction):
                    action.setVisible(action.has_any_tag(tags))

            menu_vis = any(action.isVisible() for action in menu.actions())
            menu.menuAction().setVisible(menu_vis)

        search_str = search_string or ''
        split = search_str.split()
        if not split:
            qtutils.recursively_set_menu_actions_visibility(menu=self, state=True)
            return
        elif len(split) > 1:
            _recursive_search_by_tags(menu=self, tags=split)
            return

        _recursive_search(menu=self, search_str=split[0])

    def _init_search_edit(self):
        """
        Internal function that adds a QLineEdit as the first action in the menu
        """

        self._search_action = QWidgetAction(self)
        self._search_action.setObjectName('SearchAction')
        self._search_edit = QLineEdit(self)
        self._search_edit.setPlaceholderText('Search ...')
        self._search_edit.textChanged.connect(self._on_update_search)
        self._search_action.setDefaultWidget(self._search_edit)
        self.addAction(self._search_action)
        self.addSeparator()

    def _on_update_search(self, search_string):
        """
        Internal callback function that is called when the user interacts with the search line edit
        """

        self.update_search(search_string)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains implementation for card widgets
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import *
from Qt.QtWidgets import *

from tpDcc.libs.qt.core import base, mixin, theme
from tpDcc.libs.qt.widgets import label, avatar, buttons, dividers


@mixin.theme_mixin
@mixin.cursor_mixin
class BaseCard(base.BaseWidget, object):
    def __init__(self, title=None, image=None, size=None, extra=None, type=None, parent=None):

        self._title = title
        self._size = size
        self._image = image
        self._extra = extra

        super(BaseCard, self).__init__(parent=parent)

        self.setAttribute(Qt.WA_StyledBackground)
        self.setProperty('border', False)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def get_main_layout(self):
        main_layout = QVBoxLayout()
        main_layout.setSpacing(0)
        main_layout.setContentsMargins(1, 1, 1, 1)

        return main_layout

    def ui(self):
        super(BaseCard, self).ui()

        widget_theme = self.theme()

        map_label = {
            widget_theme.small if widget_theme else theme.Theme.Sizes.SMALL: (label.BaseLabel.Levels.H4, 10),
            widget_theme.medium if widget_theme else theme.Theme.Sizes.MEDIUM: (label.BaseLabel.Levels.H3, 15),
            widget_theme.large if widget_theme else theme.Theme.Sizes.LARGE: (label.BaseLabel.Levels.H2, 20)
        }
        size = self._size or widget_theme.default if widget_theme else theme.Theme.Sizes.MEDIUM
        padding = map_label.get(size)[-1]

        self._title_layout = QHBoxLayout()
        self._title_layout.setContentsMargins(padding, padding, padding, padding)
        self._title_label = label.BaseLabel(text=self._title, parent=self)
        self._title_label.level = map_label.get(size)[0]
        if self._image:
            self._title_icon = avatar.Avatar()
            self._title_icon.image = self._image
            self._title_icon.theme_size = size
            self._title_layout.addWidget(self._title_icon)
        self._title_layout.addWidget(self._title_label)
        self._title_layout.addStretch()
        if self._extra:
            self._extra_button = buttons.BaseToolButton(parent=self).image('more').icon_only()
            self._title_layout.addWidget(self._extra_button)

        self._content_layout = QVBoxLayout()

        if self._title:
            self.main_layout.addLayout(self._title_layout)
            self.main_layout.addWidget(dividers.Divider())
        self.main_layout.addLayout(self._content_layout)

    # =================================================================================================================
    # BASE
    # =================================================================================================================

    def get_more_button(self):
        """
        Returns more button
        :return: buttons.BaseToolButton
        """

        return self._extra_button

    def set_widget(self, widget):
        """
        Adds a new widget to the card contents layout
        :param widget: QWidget
        """

        self._content_layout.addWidget(widget)

    def border(self):
        """
        Enables card border style
        :return: self
        """

        self.setProperty('border', True)
        self.style().polish(self)

        return self


@mixin.cursor_mixin
class MetaCard(base.BaseWidget, object):
    def __init__(self, extra=False, parent=None):

        self._extra = extra

        super(MetaCard, self).__init__(parent=parent)

        self.setAttribute(Qt.WA_StyledBackground)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def get_main_layout(self):
        main_layout = QVBoxLayout()
        main_layout.setSpacing(0)
        main_layout.setContentsMargins(1, 1, 1, 1)

        return main_layout

    def ui(self):
        super(MetaCard, self).ui()

        self._title_layout = QHBoxLayout()
        self._cover_label = QLabel()
        self._cover_label.setFixedSize(QSize(200, 200))
        self._avatar = avatar.Avatar()
        self._title_label = label.BaseLabel().h4()
        self._description_label = label.BaseLabel().secondary()
        self._description_label.setWordWrap(True)
        self._description_label.theme_elide_mode = Qt.ElideRight
        self._extra_btn = buttons.BaseToolButton(parent=self).image('more').icon_only()
        self._title_layout.addWidget(self._title_label)
        self._title_layout.addStretch()
        self._title_layout.addWidget(self._extra_btn)
        self._extra_btn.setVisible(self._extra)

        content_lyt = QFormLayout()
        content_lyt.setContentsMargins(5, 5, 5, 5)
        content_lyt.addRow(self._avatar, self._title_layout)
        content_lyt.addRow(self._description_label)
        self._btn_lyt = QHBoxLayout()

        self.main_layout.addWidget(self._cover_label)
        self.main_layout.addLayout(content_lyt)
        self.main_layout.addLayout(self._btn_lyt)
        self.main_layout.addStretch()

    # =================================================================================================================
    # BASE
    # =================================================================================================================

    def get_more_button(self):
        """
        Returns more button
        :return: buttons.BaseToolButton
        """

        return self._extra_button

    def setup_data(self, data_dict):
        if data_dict.get('title'):
            self._title_label.setText(data_dict.get('title'))
            self._title_label.setVisible(True)
        else:
            self._title_label.setVisible(False)

        if data_dict.get('description'):
            self._description_label.setText(data_dict.get('description'))
            self._description_label.setVisible(True)
        else:
            self._description_label.setVisible(False)

        if data_dict.get('avatar'):
            self._avatar.image = data_dict.get('avatar')
            self._avatar.setVisible(True)
        else:
            self._avatar.setVisible(False)

        if data_dict.get('cover'):
            fixed_height = self._cover_label.width()
            self._cover_label.setPixmap(data_dict.get('cover').scaledToWidth(fixed_height, Qt.SmoothTransformation))
            self._cover_label.setVisible(True)
        else:
            self._cover_label.setVisible(False)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains different graphic effects
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import *
from Qt.QtWidgets import *

from tpDcc.libs.python import decorators


class OpacityEffect(QGraphicsEffect, object):
    def __init__(self, target, duration, parent=None):
        super(OpacityEffect, self).__init__(parent=parent)

        self._target = target
        self._duration = duration
        self._animation = QPropertyAnimation(self._target, "opacity")
        self._animation.setStartValue(0.0)
        self._animation.setEndValue(0.0)

    def get_duration(self):
        return self._duration

    def set_duration(self, value):
        self._duration = value

    def get_target(self):
        return self._target

    def set_target(self, value):
        self._target = value

    def get_animation(self):
        return self._animation

    duration = property(get_duration, set_duration)
    target = property(get_target, set_target)
    animation = property(get_animation)

    def fade_in_out(self):
        """
        Executes the animation
        """

        self._animation.stop()
        self._animation.setDuration(self._duration)
        self._animation.setEasingCurve(QEasingCurve.InOutQuad)
        self._animation.setStartValue(0.0)
        self._animation.setEndValue(0.0)
        self._animation.setKeyValueAt(0.3, 1.0)
        self._animation.setKeyValueAt(0.6, 1.0)
        self._animation.start()

    # region Functions
    def fade_in(self):
        """
        Fade in the opacity property of the effect target
        """

        self._animation.stop()
        self._animation.setEasingCurve(QEasingCurve.InOutQuad)
        self._animation.setDuration(self._duration)
        self._animation.setStartValue(0)
        self._animation.setEndValue(1)
        self._animation.start()

    def fade_out(self):
        """
        Fade out the opacity property of the effect target
        """

        self._animation.stop()
        self._animation.setEasingCurve(QEasingCurve.InOutQuad)
        self._animation.setDuration(self._duration)
        self._animation.setStartValue(1)
        self._animation.setEndValue(0)
        self._animation.start()


class GraphicsLayeredBlurEffect(QGraphicsBlurEffect, object):
    def __init__(self, inner_radius=0, outer_radius=0, parent=None):
        super(GraphicsLayeredBlurEffect, self).__init__(parent=parent)

        self._inner_radius = inner_radius
        self._outer_radius = outer_radius

    @decorators.returns(float)
    def get_inner_radius(self):
        return self._inner_radius

    @decorators.accepts(float)
    def set_inner_radius(self, value):
        self._inner_radius = value

    @decorators.returns(float)
    def get_outer_radius(self):
        return self._outer_radius

    @decorators.accepts(float)
    def set_outer_radius(self, value):
        self._outer_radius = value

    inner_radius = property(get_inner_radius, set_inner_radius)
    outer_radius = property(get_outer_radius, set_outer_radius)

    def draw(self, painter):
        if self._outer_radius > 0:
            self.setBlurRadius(self.outer_radius)
            super(GraphicsLayeredBlurEffect, self).draw(painter)
        if self._inner_radius > 0:
            self.setBlurRadius(self._inner_radius)
            super(GraphicsLayeredBlurEffect, self).draw(painter)
        super(GraphicsLayeredBlurEffect, self).drawSource(painter)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains implementation for custom PySide/PyQt windows
"""

from HTMLParser import HTMLParser

from Qt.QtWidgets import *
from Qt.QtGui import *

from tpDcc.libs.qt.widgets import gif


class WidgetsFromTextParser(HTMLParser, object):
    def __init__(self, text, root_tag):
        super(WidgetsFromTextParser, self).__init__()

        self._widgets = list()
        self._root_tag = root_tag

        self._constructed = ''
        self._font = QFont('sans')

        self.feed(text)

    @property
    def widgets(self):
        return self._widgets

    def feed(self, data):
        super(WidgetsFromTextParser, self).feed(data)

        self.add_label_from_constructed()

    def handle_starttag(self, tag, attrs):
        starttag = self.get_starttag_text()
        self._constructed += starttag

        if tag == self._root_tag:
            if attrs[0][0] == 'gig':
                rem = len(starttag)
                self._constructed = self._constructed[:-rem]
                self.add_label_from_constructed()
                self.add_gif_widget(attrs[0][1])

    def handle_endtag(self, tag):
        if tag == self._root_tag:
            return

        self._constructed += "</{}>".format(tag)

    def handle_data(self, data):
        self._constructed += data

    def add_gif_widget(self, gif_file):
        gif_widget = gif.GifLabel(gif_file)
        self._widgets.append(gif_widget)

    def add_label_from_constructed(self):
        label = QLabel(self._constructed)
        label.setOpenExternalLinks(True)
        label.setWordWrap(True)
        label.setFont(self._font)
        self._constructed = ''

        self._widgets.append(label)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Class that creates a status widgets which can be used the state of an app
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import *
from Qt.QtWidgets import *
from Qt.QtGui import *

import tpDcc


class StatusWidget(QFrame, object):

    DEFAULT_DISPLAY_TIME = 10000  # milliseconds -> 15 seconds

    def __init__(self, *args):
        super(StatusWidget, self).__init__(*args)

        self._blocking = False
        self._timer = QTimer(self)

        self.setObjectName('StatusWidget')
        self.setFrameShape(QFrame.NoFrame)
        self.setFixedHeight(19)
        self.setMinimumWidth(5)

        self._label = QLabel('', self)
        self._label.setStyleSheet('background-color: transparent;')
        self._label.setCursor(Qt.IBeamCursor)
        self._label.setTextInteractionFlags(Qt.TextSelectableByMouse)
        self._label.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)

        self._button = QPushButton(self)
        self._button.setMaximumSize(QSize(17, 17))
        self._button.setIconSize(QSize(20, 20))
        self._button.hide()

        self.main_layout = QHBoxLayout(self)
        self.main_layout.setContentsMargins(5, 0, 0, 0)

        self.main_layout.addWidget(self._button)
        self.main_layout.addWidget(self._label)

        self.setLayout(self.main_layout)

        self._timer.timeout.connect(self._reset)

    def is_blocking(self):
        """
        Returns True if the status widget is blocking, otherwise return False
        :return: bool
        """

        return self._blocking

    def show_ok_message(self, message, msecs=None):
        """
        Set an ok message to be displayed in the status widget
        :param message: str
        :param msecs: int
        """

        if self.is_blocking():
            return

        icon = tpDcc.ResourcesMgr().icon('ok', extension='png')
        self._show_message(message, icon, msecs)

    def show_info_message(self, message, msecs=None):
        """
        Set an info message to be displayed in the status widget
        :param message: str
        :param msecs: int
        """

        if self.is_blocking():
            return

        icon = tpDcc.ResourcesMgr().icon('info', extension='png')
        self._show_message(message, icon, msecs)

    def show_warning_message(self, message, msecs=None):
        """
       Set a warning message to be displayed in the status widget
       :param message: str
       :param msecs: int
       """

        if self.is_blocking():
            return

        icon = tpDcc.ResourcesMgr().icon('warning', extension='png')
        self._show_message(message, icon, msecs)

    def show_error_message(self, message, msecs=None):
        """
       Set an error message to be displayed in the status widget
       :param message: str
       :param msecs: int
       """

        icon = tpDcc.ResourcesMgr().icon('error', extension='png')
        self._show_message(message, icon, msecs)

    def _reset(self):
        """
        Called when the current animation has finished
        """

        self._timer.stop()
        self._button.hide()
        self._label.setText('')
        icon = tpDcc.ResourcesMgr().icon('blank')
        self._button.setIcon(icon) if icon else self._button.setIcon(QIcon())
        self.setStyleSheet('')
        self._blocking = False

    def _show_message(self, message, icon, msecs=None, blocking=False):
        """
        Set the given text to be displayed in the status widget
        :param message: str
        :param icon: QIcon
        :param msecs: int
        :param blocking: bool
        """

        msecs = msecs or self.DEFAULT_DISPLAY_TIME
        self._blocking = blocking

        self._button.setStyleSheet('border: 0px;')

        if icon:
            self._button.setIcon(icon)
            self._button.show()
        else:
            self._button.hide()

        if message:
            self._label.setText(str(message))
            self._timer.stop()
            self._timer.start(msecs)
        else:
            self._reset()

        self.update()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains code/script related widgets
"""

from __future__ import print_function, division, absolute_import

import re
import sys
import string

from Qt.QtCore import *
from Qt.QtWidgets import *
from Qt.QtGui import *

import tpDcc as tp
from tpDcc.libs import qt
from tpDcc.libs.python import python, fileio, code, folder as folder_utils, path as path_utils
from tpDcc.libs.qt.core import qtutils


class PythonCompleter(QCompleter, object):
    def __init__(self):
        super(PythonCompleter, self).__init__()

        self._info = None
        self._file_path = None
        self._model_strings = list()
        self._reset_list = True
        self._string_model = QStringListModel(self._model_strings, self)

        self._refresh_completer = True
        self._sub_activated = False
        self._last_imports = None
        self._last_lines = None
        self._last_path = None
        self._current_defined_imports = None
        self._last_path_and_part = None
        self._current_sub_functions = None
        self._last_column = 0

        self.setCompletionMode(self.PopupCompletion)
        self.setCaseSensitivity(Qt.CaseInsensitive)
        self.setModel(self._string_model)
        self.setWrapAround(False)

        self.activated.connect(self._on_insert_completion)

    def setWidget(self, widget):
        super(PythonCompleter, self).setWidget(widget)
        self.setParent(widget)

    def keyPressEvent(self):
        return

    def show_info_popup(self, info=None):
        self._info = QTextEdit()
        self._info.setEnabled(False)
        self._info.setWindowFlags(Qt.Popup)
        self._info.show()

    def get_imports(self, paths=None):
        imports = self._get_available_modules(paths=paths)
        imports.sort()

        return imports

    def get_sub_imports(self, path):
        """
        Returns namespace in a module
        :param path: str
        :return: str
        """

        defined = code.get_defined(path)
        defined.sort()

        return defined

    def clear_completer_list(self):
        self._string_model.setStringList([])

    def text_under_cursor(self):
        cursor = self.widget().textCursor()
        cursor.select(cursor.LineUnderCursor)

        return cursor.selectedText()

    def set_filepath(self, file_path):
        if not file_path:
            return

        self._file_path = file_path

    def handle_text(self, text):
        """
        Parses a single line of text
        :param text: str
        :return: bool
        """

        if not text:
            return False

        cursor = self.widget().textCursor()
        column = cursor.columnNumber() - 1
        if column < self._last_column:
            self._last_column = column
            return False
        self._last_column = column

        if column == 1:
            return False
        text = str(text)
        passed = self.handle_from_import(text, column)
        if passed:
            return True
        passed = self.handle_sub_import(text, column)
        if passed:
            return True
        passed = self.handle_import_load(text, cursor)
        if passed:
            return True

        return False

    def handle_import(self, text):
        m = re.search(r'(from|import)(?:\s+?)(\w*)', text)
        if m:
            # TODO: Find available modules in Python path
            pass

    def handle_sub_import(self, text, column):
        m = re.search(r'(from|import)(?:\s+?)(\w*.?\w*)\.(\w*)$', text)
        if m:
            if column < m.end(2):
                return False
            from_module = m.group(2)
            module_path = code.get_package_path_from_name(from_module)
            last_part = m.group(3)
            if module_path:
                defined = self.get_imports(module_path)
                self._string_model.setStringList(defined)
                self.setCompletionPrefix(last_part)
                self.popup().setCurrentIndex(self.completionModel().index(0, 0))
                return True

        return False

    def handle_import_load(self, text, cursor):
        column = cursor.columnNumber() - 1
        text = text[:cursor.columnNumber()]
        m = re.search(r'\s*([a-zA-Z0-9._]+)\.([a-zA-Z0-9_]*)$', text)
        block_number = cursor.blockNumber()
        line_number = block_number + 1
        all_text = self.widget().toPlainText()
        scope_text = all_text[:(cursor.position() - 1)]

        if m and m.group(2):
            scope_text = all_text[:(cursor.position() - len(m.group(2)) + 1)]
        if not m:
            return False

        assignment = m.group(1)
        if column < m.end(1):
            return False

        sub_m = re.search(r'(from|import)\s+(%s)' % assignment, text)
        if sub_m:
            return False

        path = None
        sub_part = None
        target = None

        text = self.widget().toPlainText()
        lines = fileio.get_text_lines(text)

        # Search for assignments
        assign_map = code.get_ast_assignment(scope_text, line_number - 1, assignment)
        if assign_map:
            if assignment in assign_map:
                target = assign_map[assignment]
            else:
                split_assignment = assignment.split('.')
                inc = 1

                while assignment not in assign_map:
                    sub_assignment = string.join(split_assignment[:(inc * -1)], '.')
                    if sub_assignment in assign_map:
                        target = assign_map[sub_assignment]
                        break
                    inc += 1
                    if inc > (len(split_assignment) - 1):
                        break
                sub_part = string.join(split_assignment[inc:], '.')

        module_name = m.group(1)
        if target and len(target) == 2:
            if target[0] == 'import':
                module_name = target[1]
            if not target[0] == 'import':
                module_name = target[0]
                sub_part = target[1]

        # import from module
        if module_name:
            imports = None
            if lines == self.last_lines:
                imports = self.last_imports
            if not imports:
                imports = code.get_line_imports(lines)

            self._last_imports = imports
            self._last_lines = lines

            if module_name in imports:
                path = imports[module_name]
            if module_name not in imports:
                split_assignment = module_name.split('.')
                last_part = split_assignment[-1]
                if last_part in imports:
                    path = imports[last_part]

            if path and not sub_part:
                test_text = ''
                defined = None
                if path == self.last_path:
                    defined = self.current_defined_imports
                if len(m.groups()) > 0:
                    test_text = m.group(2)
                if not defined:
                    defined = self.get_imports(path)
                    if defined:
                        self._current_defined_imports = defined
                    else:
                        defined = self.get_sub_imports(path)

                custom_defined = self.custom_import_load(assign_map, module_name)
                if custom_defined:
                    defined = custom_defined
                if defined:
                    if test_text and test_text[0].islower():
                        defined.sort(key=str.swapcase)
                    self._string_model.setStringList(defined)
                    self.setCompletionPrefix(test_text)
                    self.setCaseSensitivity(Qt.CaseInsensitive)
                    self.popup().setCurrentIndex(self.completionModel().index(0, 0))
                    return True

            # import from a class of a module
            if path and sub_part:

                sub_functions = None
                if self.last_path_and_part:
                    if path == self.last_path_and_part[0] and sub_part == self.last_path_and_part[1]:
                        sub_functions = self.current_sub_functions

                if not sub_functions:
                    sub_functions = code.get_ast_class_sub_functions(path, sub_part)
                    if sub_functions:
                        self._current_sub_functions = sub_functions

                self._last_path_and_part = [path, sub_part]
                if not sub_functions:
                    return False

                test_text = ''
                if len(m.groups()) > 0:
                    test_text = m.group(2)
                if test_text and test_text[0].islower():
                    sub_functions.sort(key=str.swapcase)
                self._string_model.setStringList(sub_functions)
                self.setCompletionPrefix(test_text)
                self.setCaseSensitivity(Qt.CaseInsensitive)
                self.popup().setCurrentIndex(self.completionModel().index(0, 0))
                return True

        module_name = m.group(1)
        if module_name:
            custom_defined = self.custom_import_load(assign_map, module_name)

            test_text = ''
            if len(m.groups()) > 0:
                test_text = m.group(2)

            if test_text and test_text[0].islower():
                custom_defined.sort(key=str.swapcase)
            self._string_model.setStringList(custom_defined)
            self.setCompletionPrefix(test_text)
            self.setCaseSensitivity(Qt.CaseInsensitive)
            self.popup().setCurrentIndex(self.completionModel().index(0, 0))
            return True

        return False

    def handle_from_import(self, text, column):
        m = re.search(r'(from)(?:\s+?)(\w*.?\w*)(?:\s+?)(import)(?:\s+?)(\w+)?$', text)
        if m:
            if column < m.end(3):
                return False
            from_module = m.group(2)
            module_path = code.get_package_path_from_name(from_module)

            last_part = m.group(4)
            if not last_part:
                last_part = ''
            if module_path:
                defined = self.get_imports(module_path)
                self._string_model.setStringList(defined)
                self.setCompletionPrefix(last_part)
                self.popup().setCurrentIndex(self.completionModel().index(0, 0))

                return True

        return False

    def custom_import_load(self, assign_map, moduel_name):
        return

    def _get_available_modules(self, paths=None):
        imports = list()
        if not paths:
            paths = sys.path
        if paths:
            paths = python.force_list(paths)

        for path in paths:
            fix_path = path_utils.normalize_path(path)
            stuff_in_folder = folder_utils.get_files_and_folders(fix_path)
            for file_or_folder in stuff_in_folder:
                folder_path = path_utils.join_path(fix_path, file_or_folder)
                files = folder_utils.get_files_with_extension('py', folder_path, full_path=False)
                if '__init__.py' in files:
                    imports.append(str(file_or_folder))

            python_files = folder_utils.get_files_with_extension('py', fix_path, full_path=False)
            for python_file in python_files:
                if python_file.startswith('__'):
                    continue
                python_file_name = python_file.split('.')[0]
                imports.append(str(python_file_name))

        if imports:
            imports = list(set(imports))

        return imports

    def _on_insert_completion(self, completion_string):
        widget = self.widget()
        cursor = widget.textCursor()
        if completion_string == self.completionPrefix():
            return
        extra = len(self.completionPrefix())
        cursor.movePosition(QTextCursor.Left, cursor.KeepAnchor, extra)
        cursor.removeSelectedText()
        cursor.insertText(completion_string)
        widget.setTextCursor(cursor)


def get_syntax_format(color=None, style=''):
    """
    Returns a QTextCharFormat with the given attributes.
    """

    _color = None

    if type(color) == str:
        _color = QColor()
        _color.setNamedColor(color)
    if type(color) == list:
        _color = QColor(*color)

    if color == 'green':
        _color = Qt.green

    _format = QTextCharFormat()

    if _color:
        _format.setForeground(_color)
    if 'bold' in style:
        _format.setFontWeight(QFont.Bold)
    if 'italic' in style:
        _format.setFontItalic(True)

    return _format


def syntax_styles(name):
    if name == 'keyword':
        if tp.is_maya():
            return get_syntax_format('green', 'bold')
        if not tp.is_maya():
            return get_syntax_format([0, 150, 150], 'bold')
    if name == 'operator':
        if tp.is_maya():
            return get_syntax_format('gray')
        if not tp.is_maya():
            return get_syntax_format('darkGray')
    if name == 'brace':
        if tp.is_maya():
            return get_syntax_format('lightGray')
        if not tp.is_maya():
            return get_syntax_format('darkGray')
    if name == 'defclass':
        if tp.is_maya():
            return get_syntax_format(None, 'bold')
        if not tp.is_maya():
            return get_syntax_format(None, 'bold')
    if name == 'string':
        if tp.is_maya():
            return get_syntax_format([230, 230, 0])
        if not tp.is_maya():
            return get_syntax_format('blue')
    if name == 'string2':
        if tp.is_maya():
            return get_syntax_format([230, 230, 0])
        if not tp.is_maya():
            return get_syntax_format('lightGreen')
    if name == 'comment':
        if tp.is_maya():
            return get_syntax_format('red')
        if not tp.is_maya():
            return get_syntax_format('red')
    if name == 'self':
        if tp.is_maya():
            return get_syntax_format(None, 'italic')
        if not tp.is_maya():
            return get_syntax_format('black', 'italic')
    if name == 'bold':
        return get_syntax_format(None, 'bold')
    if name == 'numbers':
        if tp.is_maya():
            return get_syntax_format('cyan')
        if not tp.is_maya():
            return get_syntax_format('brown')


class PythonHighlighter(QSyntaxHighlighter):
    """
    Syntax highlighter for the Python language.
    """

    # Python keywords
    keywords = [
        'and', 'assert', 'break', 'class', 'continue', 'def',
        'del', 'elif', 'else', 'except', 'exec', 'finally',
        'for', 'from', 'global', 'if', 'import', 'in',
        'is', 'lambda', 'not', 'or', 'pass', 'print',
        'raise', 'return', 'try', 'while', 'yield',
        'None', 'True', 'False', 'process', 'show'
    ]

    if tp.is_maya():
        keywords += ['cmds', 'pm', 'mc', 'pymel']

    # Python operators
    operators = [
        '=',
        # Comparison
        '==', '!=', '<', '<=', '>', '>=',
        # Arithmetic
        r'\+', '-', r'\*', '/', '//', r'\%', r'\*\*',
        # In-place
        r'\+=', '-=', r'\*=', '/=', r'\%=',
        # Bitwise
        r'\^', r'\|', r'\&', r'\~', '>>', '<<',
    ]

    # Python braces
    braces = [
        r'\{', r'\}', r'\(', r'\)', r'\[', r'\]',
    ]

    def __init__(self, document):
        super(PythonHighlighter, self).__init__(document)

        # Multi-line strings (expression, flag, style)
        # FIXME: The triple-quotes in these two lines will mess up the
        # syntax highlighting from this point onward
        self.tri_single = (QRegExp("'''"), 1, syntax_styles('string2'))
        self.tri_double = (QRegExp('"""'), 2, syntax_styles('string2'))

        rules = []

        # Keyword, operator, and brace rules
        rules += [(r'\b%s\b' % w, 0, syntax_styles('keyword'))
                  for w in PythonHighlighter.keywords]
        rules += [(r'%s' % o, 0, syntax_styles('operator'))
                  for o in PythonHighlighter.operators]
        rules += [(r'%s' % b, 0, syntax_styles('brace'))
                  for b in PythonHighlighter.braces]

        # All other rules
        rules += [
            # 'self'
            (r'\bself\b', 0, syntax_styles('self')),

            # Double-quoted string, possibly containing escape sequences
            (r'"[^"\\]*(\\.[^"\\]*)*"', 0, syntax_styles('string')),
            # Single-quoted string, possibly containing escape sequences
            (r"'[^'\\]*(\\.[^'\\]*)*'", 0, syntax_styles('string')),

            # 'def' followed by an identifier
            # (r'\bdef\b\s*(\w+)', 0, syntax_styles('defclass')),
            # 'class' followed by an identifier
            # (r'\bclass\b\s*(\w+)', 0, syntax_styles('defclass')),

            # From '#' until a newline
            (r'#[^\n]*', 0, syntax_styles('comment')),
            # ('\\b\.[a-zA-Z_]+\\b(?=\()', 0, syntax_styles('bold')),
            # Numeric literals
            (r'\b[+-]?[0-9]+[lL]?\b', 0, syntax_styles('numbers')),
            (r'\b[+-]?0[xX][0-9A-Fa-f]+[lL]?\b', 0, syntax_styles('numbers')),
            (r'\b[+-]?[0-9]+(?:\.[0-9]+)?(?:[eE][+-]?[0-9]+)?\b', 0, syntax_styles('numbers')),
        ]

        # Build a QRegExp for each pattern
        self.rules = [(QRegExp(pat), index, fmt) for (pat, index, fmt) in rules]

    def highlightBlock(self, text):
        """
        Apply syntax highlighting to the given block of text.
        """

        # Do other syntax formatting
        for expression, nth, format_value in self.rules:
            index = expression.indexIn(text, 0)

            while index >= 0:
                # We actually want the index of the nth match
                index = expression.pos(nth)
                length = len(expression.cap(nth))
                self.setFormat(index, length, format_value)
                index = expression.indexIn(text, index + length)

        self.setCurrentBlockState(0)

        # Do multi-line strings
        in_multiline = self.match_multiline(text, *self.tri_single)
        if not in_multiline:
            in_multiline = self.match_multiline(text, *self.tri_double)

    def match_multiline(self, text, delimiter, in_state, style):
        """
        Do highlighting of multi-line strings. ``delimiter`` should be a
        ``QRegExp`` for triple-single-quotes or triple-double-quotes, and
        ``in_state`` should be a unique integer to represent the corresponding
        state changes when inside those strings. Returns True if we're still
        inside a multi-line string when this function is finished.
        """

        # If inside triple-single quotes, start at 0
        if self.previousBlockState() == in_state:
            start = 0
            add = 0
        # Otherwise, look for the delimiter on this line
        else:
            start = delimiter.indexIn(text)
            # Move past this match
            add = delimiter.matchedLength()

        # As long as there's a delimiter match on this line...
        while start >= 0:
            # Look for the ending delimiter
            end = delimiter.indexIn(text, start + add)
            # Ending delimiter on this line?
            if end >= add:
                length = end - start + add + delimiter.matchedLength()
                self.setCurrentBlockState(0)
            # No; multi-line string
            else:
                self.setCurrentBlockState(in_state)
                length = len(text) - start + add
            # Apply formatting
            self.setFormat(start, length, style)
            # Look for the next match
            start = delimiter.indexIn(text, start + length)

        # Return True if still inside a multi-line string, False otherwise
        if self.currentBlockState() == in_state:
            return True
        else:
            return False


class CodeLineNumber(QWidget, object):
    def __init__(self, code_editor):
        super(CodeLineNumber, self).__init__()

        self.setParent(code_editor)
        self._code_editor = code_editor

    def sizeHint(self):
        return QSize(self._code_editor._line_number_width(), 0)

    def paintEvent(self, event):
        self._code_editor._line_number_paint(event)


class CodeCompleter(PythonCompleter, object):
    def __init__(self):
        super(CodeCompleter, self).__init__()

    def keyPressEvent(self):
        return

    def _on_insert_completion(self, completion_string):
        super(CodeCompleter, self)._on_insert_completion(completion_string)

        # This stops Maya from entering edit mode in the outliner, if something is selected
        if tp.is_maya():
            tp.Dcc.focus('modelPanel1')

    def _format_live_function(self, function_instance):
        function_name = None
        if hasattr(function_instance, 'im_func'):
            args = function_instance.im_func.func_code.co_varnames
            count = function_instance.im_func.func_code.co_argcount
            args_name = ''
            if args:
                args = args[:count]
                if args[0] == 'self':
                    args = args[1:]
                args_name = string.join(args, ',')
            function_name = '{}({})'.format(function_instance.im_func.func.name, args_name)

        return function_name

    def custom_import_load(self, assign_map, module_name):

        found = list()

        if module_name == 'cmds':
            if assign_map:
                if module_name in assign_map:
                    return []
            if tp.is_maya():
                import maya.cmds as cmds
                functions = dir(cmds)
                return functions

        return found


class CodeTextEdit(QPlainTextEdit, object):

    save = Signal(object)
    saveDone = Signal(object)
    fileSet = Signal()
    findOpened = Signal(object)
    codeTextSizeChanged = Signal(object)
    mousePressed = Signal(object)

    def __init__(self, settings=None, parent=None):

        self._settings = settings
        self._text_edit_globals = dict()
        self._file_path = None
        self._last_modified = None
        self._find_widget = None
        self._completer = None
        self._skip_focus = False

        self._option_object = None

        super(CodeTextEdit, self).__init__(parent=parent)

        self.setFont(QFont('Courier', 9))
        self.setWordWrapMode(QTextOption.NoWrap)

        save_shortcut = QShortcut(QKeySequence(self.tr('Ctlr+s')), self)
        find_shortcut = QShortcut(QKeySequence(self.tr('Ctrl+f')), self)
        goto_line_shortcut = QShortcut(QKeySequence(self.tr('Ctrl+g')), self)
        duplicate_line_shortcut = QShortcut(QKeySequence(self.tr('Ctrl+d')), self)
        zoom_in_shortcut = QShortcut(QKeySequence(Qt.CTRL + Qt.Key_Plus), self)
        zoom_in_other_shortcut = QShortcut(QKeySequence(Qt.CTRL + Qt.Key_Equal), self)
        zoom_out_shortcut = QShortcut(QKeySequence(Qt.CTRL + Qt.Key_Minus), self)

        save_shortcut.activated.connect(self._on_save)
        find_shortcut.activated.connect(self._on_find)
        goto_line_shortcut.activated.connect(self._on_goto_line)
        duplicate_line_shortcut.activated.connect(self._on_duplicate_line)
        zoom_in_shortcut.activated.connect(self._zoom_in_text)
        zoom_in_other_shortcut.activated.connect(self._zoom_in_text)
        zoom_out_shortcut.activated.connect(self._zoom_out_text)

        self._line_numbers = CodeLineNumber(self)

        self._setup_highlighter()
        self._update_number_width(0)
        self._line_number_highlight()

        self.blockCountChanged.connect(self._on_update_number_width)
        self.updateRequest.connect(self._on_update_number_area)
        self.cursorPositionChanged.connect(self._on_cursor_position_changed)
        self.codeTextSizeChanged.connect(self._on_code_text_size_changed)

    def resizeEvent(self, event):
        super(CodeTextEdit, self).resizeEvent(event)
        rect = self.contentsRect()
        new_rect = QRect(rect.left(), rect.top(), self._line_number_width(), rect.height())
        self._line_numbers.setGeometry(new_rect)

    def mousePressEvent(self, event):
        self.mousePressed.emit(event)
        return super(CodeTextEdit, self).mousePressEvent(event)

    def keyPressEvent(self, event):

        pass_on = True
        quit_right_away = False

        if event.key() in (Qt.Key_Right, Qt.Key_Left, Qt.Key_Up, Qt.Key_Down):
            quit_right_away = True
        if quit_right_away:
            super(CodeTextEdit, self).keyPressEvent(event)
            return

        if self._completer:
            self._completer.activated.connect(self._on_activate)

            if self._completer.popup().isVisible():
                if event.key() == Qt.Key_Enter:
                    event.ignore()
                    return
                elif event.key() == Qt.Key_Return:
                    event.ignore()
                    return
                elif event.key() == Qt.Key_Escape:
                    event.ignore()
                    return
                elif event.key() == Qt.Key_Tab:
                    event.ignore()
                    return
                elif event.key() == Qt.Key_Backtab:
                    event.ignore()
                    return
            else:
                if event.key() == Qt.Key_Control or event.key() == Qt.Key_Shift:
                    event.ignore()
                    self._completer.popup().hide()
                    return
            if event.key() == Qt.Key_Control:
                event.ignore()
                self._completer.popup().hide()
                return

        if event.modifiers() and Qt.ControlModifier:
            if event.key() == Qt.Key_Enter or event.key() == Qt.Key_Return:
                self._run()
                return
        if event.key() == Qt.Key_Backtab or event.key() == Qt.Key_Tab:
            self._handle_tab(event)
            pass_on = False
        if event.key() == Qt.Key_Enter or event.key() == Qt.Key_Return:
            self._handle_enter(event)
            pass_on = False

        if pass_on:
            super(CodeTextEdit, self).keyPressEvent(event)

        if self._completer:
            text = self._completer.text_under_cursor()
            if text:
                result = self._completer.handle_text(text)
                if result:
                    rect = self.cursorRect()
                    scroll_width = self._completer.popup().verticalScrollBar().sizeHint().width()
                    width = self._completer.popup().sizeHintForColumn(0) + scroll_width
                    if width > 350:
                        width = 350
                    rect.setWidth(width)
                    self._completer.complete(rect)
                if not result:
                    self._completer.popup().hide()
                    self._completer.clear_completer_list()
                    self._completer.refresh_completer = True

    def wheelEvent(self, event):
        delta = event.delta()
        keys = event.modifiers()
        if keys == Qt.CTRL:
            if delta > 0:
                self._zoom_in_text()
            elif delta < 0:
                self._zoom_out_text()

        return super(CodeTextEdit, self).wheelEvent(event)

    def focusInEvent(self, event):
        if self._completer:
            self._completer.setWidget(self)
        super(CodeTextEdit, self).focusInEvent(event)
        if not self._skip_focus:
            self._update_request()

    def get_settings(self):
        return self._settings

    def set_settigns(self, settings):
        self._settings = settings

    def set_option_object(self, option_object):
        self._option_object = option_object

    def set_completer(self, completer):
        self._completer = completer()
        self._completer.setWidget(self)
        self._completer.set_filepath(self._file_path)

    def set_file(self, file_path):
        in_file = QFile(file_path)
        if in_file.open(QFile.ReadOnly | QFile.Text):
            text = in_file.readAll()
            self.setPlainText(text)
        self._file_path = file_path
        self._last_modified = fileio.get_last_modified_date(self._file_path)
        if self._completer:
            self._completer.set_filepath(file_path)
        self.fileSet.emit()

    def set_find_widget(self, widget):
        self._find_widget.set_widget(widget)

    def load_modification_date(self):
        self._last_modified = fileio.get_last_modified_date(self._file_path)

    def is_modified(self):
        return self.document().isModified()

    def _handle_enter(self, event):
        cursor = self.textCursor()
        current_block = cursor.block()
        cursor_position = cursor.positionInBlock()
        current_block_text = str(current_block.text())

        current_found = ''
        if not current_found:
            current_found = re.search('^ +', current_block_text)
            if current_found:
                current_found = current_found.group(0)

        indent = 0
        if current_found:
            indent = len(current_found)

        colon_position = current_block_text.find(':')
        comment_position = current_block_text.find('#')
        if colon_position > -1:
            sub_indent = 4
            if -1 < comment_position < colon_position:
                sub_indent = 0
            indent += sub_indent
        if cursor_position < indent:
            indent = (cursor_position - indent) + indent

        cursor.insertText(('\n' + ' ' * indent))

    def _handle_tab(self, event):
        cursor = self.textCursor()
        document = self.document()

        start_position = cursor.anchor()
        select_position = cursor.selectionStart()
        select_start_block = document.findBlock(select_position)
        start = select_position - select_start_block.position()
        end_position = cursor.position()
        if start_position > end_position:
            temp_position = end_position
            end_position = start_position
            start_position = temp_position

        if event.key() == Qt.Key_Tab:
            if not cursor.hasSelection():
                self.insertPlainText('    ')
                start_position += 4
                end_position = start_position
            else:
                cursor.setPosition(start_position)
                cursor.movePosition(QTextCursor.StartOfLine)
                cursor.setPosition(end_position, QTextCursor.KeepAnchor)
                text = cursor.selection().toPlainText()
                split_text = text.split('\n')
                edited = list()
                index = 0
                for text_split in split_text:
                    edited.append(self._add_tab(text_split))
                    if index == 0:
                        start_position += 4
                    end_position += 4
                    index += 1

                edited_text = string.join(edited, '\n')
                cursor.insertText(edited_text)
                self.setTextCursor(cursor)

        elif event.key() == Qt.Key_Backtab:
            if not cursor.hasSelection():
                cursor = self.textCursor()
                cursor.movePosition(QTextCursor.StartOfLine)
                cursor.movePosition(QTextCursor.Right, QTextCursor.KeepAnchor, 4)
                text = str(cursor.selection().toPlainText())
                if text and text == '    ':
                    cursor.insertText('')
                    self.setTextCursor(cursor)
                    start_position -= 4
                    end_position = start_position
            else:
                cursor.setPosition(start_position)
                cursor.movePosition(QTextCursor.StartOfLine)
                cursor.setPosition(end_position, QTextCursor.KeepAnchor)
                cursor.movePosition(QTextCursor.EndOfLine, QTextCursor.KeepAnchor)
                self.setTextCursor(cursor)
                text = str(cursor.selection().toPlainText())
                split_text = text.split('\n')
                edited = list()
                index = 0
                skip_indent = False
                for text_split in split_text:
                    new_string_value = text_split
                    if not skip_indent:
                        new_string_value = self._remove_tab(text_split)
                    if index == 0 and new_string_value == text_split:
                        skip_indent = True
                    if not skip_indent:
                        if new_string_value != text_split:
                            if index == 0:
                                offset = (start - 4) + 4
                                if offset > 4:
                                    offset = 4
                                start_position -= offset
                            end_position -= 4
                    edited.append(new_string_value)
                    index += 1

                edited_text = string.join(edited, '\n')
                cursor.insertText(edited_text)
                self.setTextCursor(cursor)

        cursor = self.textCursor()
        cursor.setPosition(start_position)
        cursor.setPosition(end_position, QTextCursor.KeepAnchor)
        self.setTextCursor(cursor)

    def _add_tab(self, string_value):
        return '    {}'.format(string_value)

    def _remove_tab(self, string_value):
        string_section = string_value[0:4]
        if string_section == '    ':
            return string_value[4:]

        return string_value

    def _set_text_size(self, value):
        font = self.font()
        font.setPixelSize(value)
        self.setFont(font)

    def _code_text_size_change(self, value):
        self._set_text_size(value)

    def _setup_highlighter(self):
        self._higlighter = PythonHighlighter(document=self.document())

    def _line_number_paint(self, event):
        paint = QPainter(self._line_numbers)
        if not tp.is_maya():
            paint.fillRect(event.rect(), Qt.lightGray)
        else:
            paint.fillRect(event.rect(), Qt.black)

        block = self.firstVisibleBlock()
        block_number = block.blockNumber()
        top = int(self.blockBoundingGeometry(block).translated(self.contentOffset()).top())
        bottom = int(top + self.blockBoundingGeometry(block).height())
        while block.isValid() and top <= event.rect().bottom():
            if block.isVisible() and bottom >= event.rect().top():
                number = block_number + 1
                if tp.is_maya():
                    paint.setPen(Qt.lightGray)
                else:
                    paint.setPen(Qt.black)
                paint.drawText(
                    0, top, self._line_numbers.width(), self.fontMetrics().height(), Qt.AlignRight, str(number))
            block = block.next()
            top = bottom
            bottom = top + self.blockBoundingRect(block).height()
            block_number += 1

    def _line_number_width(self):
        digits = 1
        max_value = max(1, self.blockCount())
        while max_value >= 10:
            max_value /= 10
            digits += 1
        space = 1 + self.fontMetrics().width('1') * digits

        return space

    def _line_number_highlight(self):
        extra_selection = QTextEdit.ExtraSelection()
        selections = [extra_selection]
        if not self.isReadOnly():
            selection = QTextEdit.ExtraSelection()
            if tp.is_maya():
                line_color = QColor(Qt.black)
            else:
                line_color = QColor(Qt.lightGray)
            selection.format.setBackground(line_color)
            selection.format.setProperty(QTextFormat.FullWidthSelection, True)
            selection.cursor = self.textCursor()
            selection.cursor.clearSelection()
            selections.append(selection)

        self.setExtraSelections(selections)

    def _update_number_width(self, value=0):
        self.setViewportMargins(self._line_number_width(), 0, 0, 0)

    def _update_number_area(self, rect, y_value):
        if y_value:
            self._line_numbers.scroll(0, y_value)
        else:
            self._line_numbers.update(0, rect.y(), self._line_number_width(), rect.height())
        if rect.contains(self.viewport().rect()):
            self._update_number_width()

    def _has_changed(self):
        if self._file_path:
            if path_utils.is_file(self._file_path):
                last_modified = fileio.get_last_modified_date(self._file_path)
                if last_modified != self._last_modified:
                    return True

        return False

    def _update_request(self):
        if not self._has_changed():
            return

        last_modified = fileio.get_last_modified_date(self._file_path)
        self._skip_focus = True
        permission = qtutils.get_permission(
            'File:\n{}\nhas changed, do you want to relaod it?'.format(path_utils.get_basename(self._file_path)), self)
        if permission:
            self.set_file(self._file_path)
        else:
            self._last_modified = last_modified

        self._skip_focus = False

    def _zoom_in_text(self):
        font = self.font()
        size = font.pointSize()
        size += 1
        font.setPointSize(size)
        self.setFont(QFont('Courier', size))

    def _zoom_out_text(self):
        font = self.font()
        size = font.pointSize()
        size -= 1
        font.setPointSize(size)
        self.setFont(QFont('Courier', size))

    def _run(self):
        raise NotImplementedError('run functionality is not implemented!')

    def _on_save(self):
        if not self.document().isModified():
            qt.logger.warning('No changes to save in {}'.format(self._file_path))
            return

        old_last_modified = self._last_modified
        try:
            self.save.emit(self)
        except Exception:
            pass

        new_last_modified = fileio.get_last_modified_date(self._file_path)
        if old_last_modified == new_last_modified:
            self.saveDone.emit(False)

        if old_last_modified != new_last_modified:
            self.saveDone.emit(True)
            self.document().setModified(False)
            self._last_modified = new_last_modified

    def _on_find(self):
        self.findOpened.emit(self)

    def _on_goto_line(self):
        line = qtutils.get_comment(parent=self, text_message='', title='Goto Line')
        if not line:
            return

        line_number = int(line)
        text_cursor = self.textCursor()
        block_number = text_cursor.blockNumber()

        number = line_number - block_number
        if number > 0:
            move_type = text_cursor.NextBlock
            number -= 2
        if number < 0:
            move_type = text_cursor.PreviousBlock
            number = abs(number)

        text_cursor.movePosition(move_type, text_cursor.MoveAnchor, (number + 1))
        self.setTextCursor(text_cursor)

    def _on_duplicate_line(self):
        text_cursor = self.textCursor()

        selected_text = text_cursor.selectedText()
        if not selected_text:
            text_cursor.beginEditBlock()
            text_cursor.movePosition(QTextCursor.StartOfLine, QTextCursor.MoveAnchor)
            text_cursor.movePosition(QTextCursor.EndOfLine, QTextCursor.KeepAnchor)
            selected_text = text_cursor.selectedText()
            text_cursor.movePosition(QTextCursor.EndOfLine, QTextCursor.MoveAnchor)
            text_cursor.insertBlock()
            text_cursor.insertText(selected_text)
            text_cursor.movePosition(QTextCursor.EndOfLine, QTextCursor.MoveAnchor)
            self.setTextCursor(text_cursor)
            text_cursor.endEditBlock()
        else:
            text_cursor.beginEditBlock()
            text_cursor.setPosition(text_cursor.selectionEnd(), QTextCursor.MoveAnchor)
            self.setTextCursor(text_cursor)
            text_cursor.insertBlock()
            position = text_cursor.position()
            text_cursor.insertText(selected_text)
            end_position = text_cursor.position()
            text_cursor.setPosition(position, QTextCursor.MoveAnchor)
            text_cursor.setPosition(end_position, QTextCursor.KeepAnchor)
            self.setTextCursor(text_cursor)
            text_cursor.endEditBlock()

    def _on_update_number_width(self, value=0):
        self.setViewportMargins(self._line_number_width(), 0, 0, 0)

    def _on_update_number_area(self, rect, y_value):
        if y_value:
            self._line_numbers.scroll(0, y_value)
        else:
            self._line_numbers.update(0, rect.y(), self._line_number_width(), rect.height())

        if rect.contains(self.viewport().rect()):
            self._update_number_width()

    def _on_cursor_position_changed(self):
        self._line_number_highlight()

    def _on_code_text_size_changed(self, value):
        self._set_text_size(value)

    def _on_activate(self):
        pass



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Tool that shows a widget with changelog info
"""

from __future__ import print_function, division, absolute_import

import os
import json
from collections import OrderedDict

from Qt.QtCore import *
from Qt.QtWidgets import *

from tpDcc.libs.qt.core import dialog
from tpDcc.libs.qt.widgets import accordion


class Changelog(dialog.Dialog, object):
    def __init__(self):
        super(Changelog, self).__init__(
            name='ChangelogDialog',
            title='Changelog'
        )

        self.setWindowFlags(self.windowFlags() ^ Qt.FramelessWindowHint)

    def ui(self):
        super(Changelog, self).ui()
        self.set_logo('changelog_logo')

        self.main_layout.setAlignment(Qt.AlignTop)
        self.main_layout.setContentsMargins(0, 0, 0, 0)

        self.setFixedWidth(600)
        self.setMaximumHeight(800)

        scroll_layout = QVBoxLayout()
        scroll_layout.setContentsMargins(2, 2, 2, 2)
        scroll_layout.setSpacing(2)
        scroll_layout.setAlignment(Qt.AlignTop)
        central_widget = QWidget()
        central_widget.setSizePolicy(QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding))
        scroll = QScrollArea()
        scroll.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        scroll.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        scroll.setWidgetResizable(True)
        scroll.setFocusPolicy(Qt.NoFocus)
        ok_btn = QPushButton('OK')
        ok_btn.clicked.connect(self.close)
        self.main_layout.addWidget(scroll)
        self.main_layout.setAlignment(Qt.AlignTop)
        self.main_layout.addWidget(ok_btn)
        scroll.setWidget(central_widget)
        central_widget.setLayout(scroll_layout)
        self.setSizePolicy(QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding))
        self.main_layout = scroll_layout

        # ===========================================================================================

        self.version_accordion = accordion.AccordionWidget(parent=self)
        self.version_accordion.rollout_style = accordion.AccordionStyle.MAYA
        self.main_layout.addWidget(self.version_accordion)

        # ===========================================================================================

        changelog_json_file = os.path.join(
            os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'changelog.json')
        if not os.path.isfile(changelog_json_file):
            return

        with open(changelog_json_file, 'r') as f:
            changelog_data = json.load(f, object_pairs_hook=OrderedDict)
        if not changelog_data:
            return

        changelog_data = OrderedDict(sorted(changelog_data.items(), reverse=True))

        for version, elements in changelog_data.items():
            self._create_version(version, elements)

        last_version_item = self.version_accordion.item_at(0)
        last_version_item.set_collapsed(False)

    def _create_version(self, version, elements):

        version_widget = QWidget()
        version_layout = QVBoxLayout()
        version_layout.setContentsMargins(0, 0, 0, 0)
        version_layout.setSpacing(0)
        version_layout.setAlignment(Qt.AlignTop)
        version_widget.setLayout(version_layout)
        self.version_accordion.add_item(version, version_widget, collapsed=True)

        version_label = QLabel()
        version_layout.addWidget(version_label)
        version_text = ''
        for item in elements:
            version_text += '- {}\n'.format(item)
        version_label.setText(version_text)

        self.main_layout.addSpacing(5)


def run():
    new_changelog = Changelog()
    new_changelog.exec_()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that defines that implements different types of actions
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import *
from Qt.QtWidgets import *


class SeparatorLine(QFrame, object):

    pass


class SeparatorWidgetAction(QFrame, object):

    pass


class SeparatorAction(QWidgetAction, object):
    def __init__(self, label='', parent=None):
        super(SeparatorAction, self).__init__(parent)

        self._widget = SeparatorWidgetAction(parent)
        self._label = QLabel(self._widget)
        self._label.setText(label)
        self._line = SeparatorLine(self._widget)
        self._line.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

    def widget(self):
        """
        Returns separator widget action object
        :return: SeparatorWidgetAction
        """

        return self._widget

    def label(self):
        """
        Returns label widget object
        :return: SeparatorLine
        """

        return self._label

    def line(self):
        """
        Returns line widget
        :return: SeparatorLine
        """

        return self._line

    def setText(self, text):
        """
        Overrides base QWidgetAction setText function
        Sets the text of the separator
        :param text: str
        """

        self._label().setText(text)

    def createWidget(self, menu):
        """
        Overrides base QWidgetAction createWidget function
        :param menu: QMenu
        """

        action_widget = self.widget()
        action_layout = QHBoxLayout(action_widget)
        action_layout.setContentsMargins(0, 0, 0, 0)
        action_layout.addWidget(self.label())
        action_layout.addWidget(self.line())
        action_widget.setLayout(action_layout)

        return action_widget


class LabelAction(QWidgetAction, object):

    def __init__(self, name='', parent=None):
        super(LabelAction, self).__init__(parent)

        self._name = name

    def createWidget(self, menu):
        """
        Overrides base QWidgetAction createWidget function
        :param menu: QMenu
        """

        widget = QFrame(self.parent())
        widget.setObjectName('filterByAction')
        title = self._name
        label = QCheckBox(widget)
        label.setText(title)
        label.setAttribute(Qt.WA_TransparentForMouseEvents)
        label.toggled.connect(self._on_triggered)
        label.setStyleSheet("""
        #QCheckBox::indicator:checked {
            image: url(none.png)
        }
        QCheckBox::indicator:unchecked {
            image: url(none.png)
        }
        """)
        action_layout = QHBoxLayout(widget)
        action_layout.setContentsMargins(0, 0, 0, 0)
        action_layout.addWidget(label, stretch=1)
        widget.setLayout(action_layout)

        return widget

    def _on_triggered(self, checked=None):
        """
        Internal callback function that is triggered when the checkbox value has changed
        :param checked: bool
        """

        self.triggered.emit()
        self.parent().close()


class SliderWidgetAction(QFrame):
    pass


class SliderAction(QWidgetAction):

    def __init__(self, label="", parent=None):
        super(SliderAction, self).__init__(parent)

        self._widget = SliderWidgetAction(parent)
        self._label = QLabel(label, self._widget)

        self._slider = QSlider(Qt.Horizontal, self._widget)
        self._slider.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        self.valueChanged = self._slider.valueChanged

    def widget(self):
        """
        Returns the widget for this action
        :return: QWidget
        """

        return self._widget

    def label(self):
        """
        Returns label widget for this action
        :return: QLabel
        """

        return self._label

    def slider(self):
        """
        Returns slider action widget for this action
        :return: SliderWidgetAction
        """

        return self._slider

    def createWidget(self, menu):
        """
        Overrides base WidgetAction createWidget function
        :param menu: QMenu
        """

        action_widget = self.widget()

        action_layout = QHBoxLayout(action_widget)
        action_layout.setContentsMargins(0, 0, 0, 0)
        action_layout.addWidget(self.label())
        action_layout.addWidget(self.slider())
        action_widget.setLayout(action_layout)

        return action_widget



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains different classes to handle Qt settings
"""

from __future__ import print_function, division, absolute_import

import os
import sys

from Qt.QtCore import *
from Qt.QtWidgets import *

from tpDcc.libs.python import settings


class QtSettings(QSettings, object):
    def __init__(self, filename, window=None, max_files=10,):
        super(QtSettings, self).__init__(filename, QSettings.IniFormat, window)

        self._max_files = max_files
        self._window = window
        if self._window:
            self._groups = [window.objectName(), 'RecentFiles']
        self._initialize()

    def has_setting(self, setting_name):
        return self.get(setting_name)

    def get(self, setting_name, default_value=None, setting_group=None):
        """
        Returns the setting stored with the given name
        :param setting_name: str
        :param default_value: variant
        :param default_value: variant
        :param setting_group: str
        :return:
        """

        if setting_group:
            setting_name = '{}/{}'.format(setting_group, setting_name)

        val = self.value(setting_name)
        if not val:
            return default_value

        return val

    def getw(self, setting_name, default_value=None):
        """
        Returns the window setting stored with the given name
        :param setting_name: str
        :param default_value: variant
        :return:
        """

        if self._window:
            val = self.value(self._window.objectName().upper() + '/' + setting_name)
            if not val:
                return default_value
        else:
            val = self.value(setting_name)
            if not val:
                return default_value

        return val

    def set(self, setting_name, setting_value):
        """
        Stores a new settings with the given name and the given value
        If the given setting already exists, it will be overwrite
        :param setting_name: str, setting name we want store
        :param setting_value: variant, setting value we want to store
        """

        self.setValue(setting_name, setting_value)

    def setw(self, setting_name, setting_value):
        """
        Stores a new window setting with the given name and the given value
        :param setting_name: str, setting name we want to store
        :param setting_value: variant, setting value we want to store
        """

        if self._window:
            self.setValue(self._window.objectName().upper() + '/' + setting_name, setting_value)
        else:
            self.setValue(setting_name, setting_value)

    def get_groups(self):
        """
        Returns the current preferences groups
        :return: list
        """

        return self._groups

    groups = property(get_groups)

    def add_group(self, group):
        """
        Add a group to the current preferences groups
        :param group: str, name of group to add
        :return: bool, True if the group was successfully added
        """

        if group not in self._groups:
            self._groups.append(group)
            return True

        return False

    def remove_group(self, group):
        """
        Remove a group from the preferences group
        :param group: str, name of group to remove
        :return: bool, True if the group was successfully removed
        """

        if group in self._groups:
            self._groups.remove(group)
            return True

        return False

    def window_keys(self):
        """
        Returns a list of all window keys to save for layouts or rebuilding last layout on launch
        :return: list
        """

        if self._window is None:
            return []

        result = [self._window.objectName()]
        for dock in self._window.findChildren(QDockWidget):
            dock_name = dock.objectName()
            result.append(str(dock_name))

        return result

    def prefs_keys(self):
        """
        Returns a list of preferences keys
        :return: list<str>, list of user prefs keys
        """

        results = list()
        self.beginGroup('Preferences')
        results = self.childKeys()
        self.endGroup()
        return results

    def get_layouts(self):
        """
        Returns a list of window layout names
        :return: list
        """

        layout_names = list()
        layout_keys = ['%s/geometry' % x for x in self.window_keys()]

        for k in self.allKeys():
            if 'geometry' in k:
                attrs = k.split('/geometry/')
                if len(attrs) > 1:
                    layout_names.append(str(attrs[-1]))

        return sorted(list(set(layout_names)))

    def save_layout(self, layout):
        """
        Save a named layout
        :param layout: str, layout name to save
        """

        sys.utils.logger.info('Saving Layout: "{}"'.format(layout))
        self.setValue(self._window.objectName() + '/geometry/%s' % layout, self._window.saveGeoemtry())
        self.setValue(self._window.objectName() + '/windowState/%s' % layout, self._window.saveState())

        for dock in self._window.findChildren(QDockWidget):
            dock_name = dock.objectName()
            self.setValue('%s/geometry/%s' % (dock_name, layout), dock.saveGeometry())

    def restore_layout(self, layout):
        """
        Restore a named layout
        :param layout: str, layout name to restores
        """

        sys.utils.logger.info('Restoring layout: "{}"'.format(layout))
        window_keys = self.window_keys()

        for widget_name in window_keys:
            key_name = '%s/geometry/%s' % (widget_name, layout)
            if widget_name != self._window.objectName():
                dock = self._window.findChildren(QDockWidget, widget_name)
                if dock:
                    if key_name in self.allKeys():
                        value = self.value(key_name)
                        dock[0].restoreGeometry(value)
            else:
                if key_name in self.allKeys():
                    value = self.value(key_name)
                    self._window.restoreGeometry(value)

                window_state = '%s/windowState/%s' % (widget_name, layout)
                if window_state in self.allKeys():
                    self._window.restoreState(self.value(window_state))

    def delete_layout(self, layout):
        """
        Delete a named layout
        :param layout: str, layout name to restore
        """

        sys.utils.logger.info('Deleting layout: "{}"'.format(layout))
        window_keys = self.window_keys()

        for widget_name in window_keys:
            key_name = '%s/geometry/%s' % (widget_name, layout)
            if key_name in self.allKeys():
                self.remove(key_name)

            if widget_name == self._window.objectName():
                window_state = '%s/windowState/%s' % (widget_name, layout)
                if window_state in self.allKeys():
                    self.remoev(window_state)

    def get_default_value(self, key, *groups):
        """
        Returns the default values for a group
        :param key: str, key to search for
        :return: variant, default value of key (None if not found)
        """

        if self.group():
            try:
                self.endGroup()
            except Exception:
                pass

        result = None

        if not groups:
            return None

        group_name = groups[0]
        for group in groups[1:]:
            group_name += '/%s' % group

        group_name += '/%s' % key
        group_name += '/%s' % 'default'

        if group_name in self.allKeys():
            result = self.value(group_name)

        return result

    def delete_file(self):
        """
        Delete the preferences file on disk
        """

        sys.utils.logger.info('Deleting Settings: "{}"'.format(self.fileName()))
        return os.remove(self.fileName())

    def get_recent_files(self):
        """
        Get a tuple of the most recent files
        """

        recent_files = list()
        cnt = self.beginReadArray('RecentFiles')
        for i in range(cnt):
            self.setArrayIndex(i)
            fn = self.value('file')
            recent_files.append(fn)
        self.endArray()

        return tuple(recent_files)

    def add_recent_file(self, filename):
        """
        Adds a recent file to the stack
        :param filename: str, file name to add
        """

        recent_files = self.get_recent_files()
        if filename in recent_files:
            recent_files = tuple(x for x in recent_files if x != filename)

        recent_files = recent_files + (filename,)
        self.beginWriteArray('RecentFiles')
        for i in range(len(recent_files)):
            self.setArrayIndex(i)
            self.setValue('file', recent_files[i])
        self.endArray()

    def clear_recent_files(self):
        self.remove('RecentFiles')

    def _initialize(self):
        if self._window:
            window_name = self._window.objectName().upper()
            if window_name not in self.childGroups():
                if self._window is not None:
                    self.setValue(window_name + '/geometry/default', self._window.saveGeometry())

                    if isinstance(self._window, QMainWindow):
                        self.setValue(window_name + '/windowState/default', self._window.saveState())

            if 'RecentFiles' not in self.childGroups():
                self.beginWriteArray('RecentFiles', 0)
                self.endArray()

            while self.group():
                self.endGroup()


class QtIniSettings(settings.INISettings, object):
    def __init__(self, filename):
        super(QtIniSettings, self).__init__(filename)

    def export_widget(self, option, widget):
        """
        Serializes a widget's state to an option, value pair Sets the value with an associated key in
        the section at the top of the section stack.

        :param str option:   The key to use in storing the value.
        :param str widget:   The QWidget to extract value to store.
        """

        if isinstance(widget, QComboBox):
            self.export_option(option, widget.currentIndex())
        elif isinstance(widget, QCheckBox):
            self.export_option(option, widget.isChecked())
        elif isinstance(widget, QToolButton):
            if widget.isCheckable():
                self.export_option(option, widget.isChecked())
        elif isinstance(widget, (QSpinBox, QDoubleSpinBox)):
            self.export_option(option, widget.value())
        elif isinstance(widget, QLineEdit):
            self.export_option(option, widget.text())
        elif isinstance(widget, (QSize, QPoint)):
            self.export_option(option, widget.toTuple())
        else:
            assert False, "Unknown control type"

    def import_widget(self, option, widget):
        """
        Serializes a widget's state to an option, value pair Sets the value with an associated key in the
        section at the top of the section stack.

        :param str option:   The key to use in storing the value.
        :param str widget:   The QWidget to extract value to store.
        """

        if isinstance(widget, QComboBox):
            d = 0 if (widget.currentIndex() < 0) else widget.currentIndex()
            v = self.import_option(option, d)
            widget.setCurrentIndex(v)
        elif isinstance(widget, QCheckBox):
            d = widget.isChecked()
            v = self.import_option(option, d)
            widget.setChecked(v)
        elif isinstance(widget, QToolButton):
            if widget.isCheckable():
                d = widget.isChecked()
                v = self.import_option(option, d)
                widget.setChecked(v)
        elif isinstance(widget, (QSpinBox, QDoubleSpinBox)):
            d = widget.value()
            v = self.import_option(option, d)
            widget.setValue(v)
        elif isinstance(widget, QLineEdit):
            d = widget.text()
            v = self.import_option(option, d)
            widget.setText(v)
        elif isinstance(widget, QSize):
            d = widget.toTuple()
            v = self.import_option(option, d, True)
            widget.setWidth(v[0])
            widget.setHeight(v[1])
        elif isinstance(widget, QPoint):
            d = widget.toTuple()
            v = self.import_option(option, d, True)
            widget.setX(v[0])
            widget.setY(v[1])
        else:
            assert False, "Unknown control type"
